import numpy as np
import pyarrow as pa
from tqdm import tqdm

from valor_lite.cache.ephemeral import MemoryCacheWriter
from valor_lite.cache.persistent import FileCacheWriter
from valor_lite.classification.annotation import Classification
from valor_lite.classification.evaluator import Builder


class Loader(Builder):
    def __init__(
        self,
        writer: MemoryCacheWriter | FileCacheWriter,
        roc_curve_writer: MemoryCacheWriter | FileCacheWriter,
        intermediate_writer: MemoryCacheWriter | FileCacheWriter,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        super().__init__(
            writer=writer,
            roc_curve_writer=roc_curve_writer,
            intermediate_writer=intermediate_writer,
            metadata_fields=metadata_fields,
        )

        # internal state
        self._labels: dict[str, int] = {}
        self._index_to_label: dict[int, str] = {}
        self._datum_count = 0

    def _add_label(self, value: str) -> int:
        idx = self._labels.get(value, None)
        if idx is None:
            idx = len(self._labels)
            self._labels[value] = idx
            self._index_to_label[idx] = value
        return idx

    def add_data(
        self,
        classifications: list[Classification],
        show_progress: bool = False,
    ):
        """
        Adds classifications to the cache.

        Parameters
        ----------
        classifications : list[Classification]
            A list of Classification objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """

        disable_tqdm = not show_progress
        for classification in tqdm(classifications, disable=disable_tqdm):
            if len(classification.predictions) == 0:
                raise ValueError(
                    "Classifications must contain at least one prediction."
                )

            # prepare metadata
            datum_metadata = (
                classification.metadata if classification.metadata else {}
            )

            # write to cache
            rows = []
            gidx = self._add_label(classification.groundtruth)
            max_score_idx = np.argmax(np.array(classification.scores))
            for idx, (plabel, score) in enumerate(
                zip(classification.predictions, classification.scores)
            ):
                pidx = self._add_label(plabel)
                rows.append(
                    {
                        # metadata
                        **datum_metadata,
                        # datum
                        "datum_uid": classification.uid,
                        "datum_id": self._datum_count,
                        # groundtruth
                        "gt_label": classification.groundtruth,
                        "gt_label_id": gidx,
                        # prediction
                        "pd_label": plabel,
                        "pd_label_id": pidx,
                        "pd_score": float(score),
                        "pd_winner": max_score_idx == idx,
                        # pair
                        "match": (gidx == pidx) and pidx >= 0,
                    }
                )
            self._writer.write_rows(rows)

            # update datum count
            self._datum_count += 1



from .annotation import Bitmask, BoundingBox, Detection, Polygon
from .evaluator import Evaluator
from .loader import Loader
from .metric import Metric, MetricType
from .shared import EvaluatorInfo

__all__ = [
    "Bitmask",
    "BoundingBox",
    "Detection",
    "Polygon",
    "Metric",
    "MetricType",
    "Loader",
    "Evaluator",
    "EvaluatorInfo",
]



from collections import defaultdict

import numpy as np
import pyarrow as pa
from numpy.typing import NDArray

from valor_lite.object_detection.metric import Metric, MetricType


def unpack_precision_recall_into_metric_lists(
    counts: NDArray[np.uint64],
    precision_recall_f1: NDArray[np.float64],
    average_precision: NDArray[np.float64],
    mean_average_precision: NDArray[np.float64],
    average_recall: NDArray[np.float64],
    mean_average_recall: NDArray[np.float64],
    pr_curve: NDArray[np.float64],
    iou_thresholds: list[float],
    score_thresholds: list[float],
    index_to_label: dict[int, str],
):
    metrics = defaultdict(list)

    metrics[MetricType.AP] = [
        Metric.average_precision(
            value=float(average_precision[iou_idx][label_idx]),
            iou_threshold=iou_threshold,
            label=label,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for label_idx, label in index_to_label.items()
    ]

    metrics[MetricType.mAP] = [
        Metric.mean_average_precision(
            value=float(mean_average_precision[iou_idx]),
            iou_threshold=iou_threshold,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.APAveragedOverIOUs] = [
        Metric.average_precision_averaged_over_IOUs(
            value=float(average_precision.mean(axis=0)[label_idx]),
            iou_thresholds=iou_thresholds,
            label=label,
        )
        for label_idx, label in index_to_label.items()
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.mAPAveragedOverIOUs] = [
        Metric.mean_average_precision_averaged_over_IOUs(
            value=float(mean_average_precision.mean()),
            iou_thresholds=iou_thresholds,
        )
    ]

    metrics[MetricType.AR] = [
        Metric.average_recall(
            value=float(average_recall[score_idx, label_idx]),
            iou_thresholds=iou_thresholds,
            score_threshold=score_threshold,
            label=label,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
        for label_idx, label in index_to_label.items()
    ]

    metrics[MetricType.mAR] = [
        Metric.mean_average_recall(
            value=float(mean_average_recall[score_idx]),
            iou_thresholds=iou_thresholds,
            score_threshold=score_threshold,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.ARAveragedOverScores] = [
        Metric.average_recall_averaged_over_scores(
            value=float(average_recall.mean(axis=0)[label_idx]),
            score_thresholds=score_thresholds,
            iou_thresholds=iou_thresholds,
            label=label,
        )
        for label_idx, label in index_to_label.items()
    ]

    # TODO - (c.zaloom) will be removed in the future
    metrics[MetricType.mARAveragedOverScores] = [
        Metric.mean_average_recall_averaged_over_scores(
            value=float(mean_average_recall.mean()),
            score_thresholds=score_thresholds,
            iou_thresholds=iou_thresholds,
        )
    ]

    metrics[MetricType.PrecisionRecallCurve] = [
        Metric.precision_recall_curve(
            precisions=pr_curve[iou_idx, label_idx, :, 0].tolist(),
            scores=pr_curve[iou_idx, label_idx, :, 1].tolist(),
            iou_threshold=iou_threshold,
            label=label,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for label_idx, label in index_to_label.items()
    ]

    for label_idx, label in index_to_label.items():
        for score_idx, score_threshold in enumerate(score_thresholds):
            for iou_idx, iou_threshold in enumerate(iou_thresholds):

                row = counts[iou_idx, score_idx, :, label_idx]
                kwargs = {
                    "label": label,
                    "iou_threshold": iou_threshold,
                    "score_threshold": score_threshold,
                }
                metrics[MetricType.Counts].append(
                    Metric.counts(
                        tp=int(row[0]),
                        fp=int(row[1]),
                        fn=int(row[2]),
                        **kwargs,
                    )
                )

                row = precision_recall_f1[iou_idx, score_idx, :, label_idx]
                metrics[MetricType.Precision].append(
                    Metric.precision(
                        value=float(row[0]),
                        **kwargs,
                    )
                )
                metrics[MetricType.Recall].append(
                    Metric.recall(
                        value=float(row[1]),
                        **kwargs,
                    )
                )
                metrics[MetricType.F1].append(
                    Metric.f1_score(
                        value=float(row[2]),
                        **kwargs,
                    )
                )

    return metrics


def unpack_confusion_matrix(
    confusion_matrices: NDArray[np.uint64],
    unmatched_groundtruths: NDArray[np.uint64],
    unmatched_predictions: NDArray[np.uint64],
    index_to_label: dict[int, str],
    iou_thresholds: list[float],
    score_thresholds: list[float],
) -> list[Metric]:
    metrics = []
    for iou_idx, iou_thresh in enumerate(iou_thresholds):
        for score_idx, score_thresh in enumerate(score_thresholds):
            cm_dict = {}
            ugt_dict = {}
            upd_dict = {}
            for idx, label in index_to_label.items():
                ugt_dict[label] = int(
                    unmatched_groundtruths[iou_idx, score_idx, idx]
                )
                upd_dict[label] = int(
                    unmatched_predictions[iou_idx, score_idx, idx]
                )
                for pidx, plabel in index_to_label.items():
                    if label not in cm_dict:
                        cm_dict[label] = {}
                    cm_dict[label][plabel] = int(
                        confusion_matrices[iou_idx, score_idx, idx, pidx]
                    )
            metrics.append(
                Metric.confusion_matrix(
                    confusion_matrix=cm_dict,
                    unmatched_ground_truths=ugt_dict,
                    unmatched_predictions=upd_dict,
                    iou_threshold=iou_thresh,
                    score_threshold=score_thresh,
                )
            )
    return metrics


def create_mapping(
    tbl: pa.Table,
    pairs: NDArray[np.float64],
    index: int,
    id_col: str,
    uid_col: str,
) -> dict[int, str]:
    col = pairs[:, index].astype(np.int64)
    values, indices = np.unique(col, return_index=True)
    indices = indices[values >= 0]
    return {
        tbl[id_col][idx].as_py(): tbl[uid_col][idx].as_py() for idx in indices
    }


def unpack_examples(
    detailed_pairs: NDArray[np.float64],
    mask_tp: NDArray[np.bool_],
    mask_fn: NDArray[np.bool_],
    mask_fp: NDArray[np.bool_],
    iou_thresholds: list[float],
    score_thresholds: list[float],
    index_to_datum_id: dict[int, str],
    index_to_groundtruth_id: dict[int, str],
    index_to_prediction_id: dict[int, str],
) -> list[Metric]:
    metrics = []
    ids = detailed_pairs[:, :5].astype(np.int64)
    unique_datums = np.unique(detailed_pairs[:, 0].astype(np.int64))
    for datum_index in unique_datums:
        mask_datum = detailed_pairs[:, 0] == datum_index
        mask_datum_tp = mask_tp & mask_datum
        mask_datum_fp = mask_fp & mask_datum
        mask_datum_fn = mask_fn & mask_datum

        datum_id = index_to_datum_id[datum_index]
        for iou_idx, iou_thresh in enumerate(iou_thresholds):
            for score_idx, score_thresh in enumerate(score_thresholds):

                unique_tp = np.unique(
                    ids[np.ix_(mask_datum_tp[iou_idx, score_idx], (0, 1, 2, 3, 4))], axis=0  # type: ignore - numpy ix_ typing
                )
                unique_fp = np.unique(
                    ids[np.ix_(mask_datum_fp[iou_idx, score_idx], (0, 2, 4))], axis=0  # type: ignore - numpy ix_ typing
                )
                unique_fn = np.unique(
                    ids[np.ix_(mask_datum_fn[iou_idx, score_idx], (0, 1, 3))], axis=0  # type: ignore - numpy ix_ typing
                )

                tp = [
                    (
                        index_to_groundtruth_id[row[1]],
                        index_to_prediction_id[row[2]],
                    )
                    for row in unique_tp
                ]
                fp = [index_to_prediction_id[row[1]] for row in unique_fp]
                fn = [index_to_groundtruth_id[row[1]] for row in unique_fn]
                metrics.append(
                    Metric.examples(
                        datum_id=datum_id,
                        true_positives=tp,
                        false_negatives=fn,
                        false_positives=fp,
                        iou_threshold=iou_thresh,
                        score_threshold=score_thresh,
                    )
                )
    return metrics


def create_empty_confusion_matrix_with_examples(
    iou_threhsold: float,
    score_threshold: float,
    index_to_label: dict[int, str],
) -> Metric:
    unmatched_groundtruths = dict()
    unmatched_predictions = dict()
    confusion_matrix = dict()
    for label in index_to_label.values():
        unmatched_groundtruths[label] = {"count": 0, "examples": []}
        unmatched_predictions[label] = {"count": 0, "examples": []}
        confusion_matrix[label] = {}
        for plabel in index_to_label.values():
            confusion_matrix[label][plabel] = {"count": 0, "examples": []}

    return Metric.confusion_matrix_with_examples(
        confusion_matrix=confusion_matrix,
        unmatched_ground_truths=unmatched_groundtruths,
        unmatched_predictions=unmatched_predictions,
        iou_threshold=iou_threhsold,
        score_threshold=score_threshold,
    )


def _unpack_confusion_matrix_with_examples(
    metric: Metric,
    ids: NDArray[np.int32],
    mask_matched: NDArray[np.bool_],
    mask_fp_unmatched: NDArray[np.bool_],
    mask_fn_unmatched: NDArray[np.bool_],
    index_to_datum_id: dict[int, str],
    index_to_groundtruth_id: dict[int, str],
    index_to_prediction_id: dict[int, str],
    index_to_label: dict[int, str],
):
    if not isinstance(metric.value, dict):
        raise TypeError("expected metric to contain a dictionary value")

    unique_matches = np.unique(
        ids[np.ix_(mask_matched, (0, 1, 2, 3, 4))], axis=0  # type: ignore - numpy ix_ typing
    )
    unique_unmatched_predictions = np.unique(
        ids[np.ix_(mask_fp_unmatched, (0, 2, 4))], axis=0  # type: ignore - numpy ix_ typing
    )
    unique_unmatched_groundtruths = np.unique(
        ids[np.ix_(mask_fn_unmatched, (0, 1, 3))], axis=0  # type: ignore - numpy ix_ typing
    )

    n_matched = unique_matches.shape[0]
    n_unmatched_predictions = unique_unmatched_predictions.shape[0]
    n_unmatched_groundtruths = unique_unmatched_groundtruths.shape[0]
    n_max = max(n_matched, n_unmatched_groundtruths, n_unmatched_predictions)

    for idx in range(n_max):
        if idx < n_unmatched_groundtruths:
            label = index_to_label[unique_unmatched_groundtruths[idx, 2]]
            metric.value["unmatched_ground_truths"][label]["count"] += 1
            metric.value["unmatched_ground_truths"][label]["examples"].append(
                {
                    "datum_id": index_to_datum_id[
                        unique_unmatched_groundtruths[idx, 0]
                    ],
                    "ground_truth_id": index_to_groundtruth_id[
                        unique_unmatched_groundtruths[idx, 1]
                    ],
                }
            )
        if idx < n_unmatched_predictions:
            label_id = unique_unmatched_predictions[idx, 2]
            label = index_to_label[label_id]
            metric.value["unmatched_predictions"][label]["count"] += 1
            metric.value["unmatched_predictions"][label]["examples"].append(
                {
                    "datum_id": index_to_datum_id[
                        unique_unmatched_predictions[idx, 0]
                    ],
                    "prediction_id": index_to_prediction_id[
                        unique_unmatched_predictions[idx, 1]
                    ],
                }
            )
        if idx < n_matched:
            glabel = index_to_label[unique_matches[idx, 3]]
            plabel = index_to_label[unique_matches[idx, 4]]
            metric.value["confusion_matrix"][glabel][plabel]["count"] += 1
            metric.value["confusion_matrix"][glabel][plabel][
                "examples"
            ].append(
                {
                    "datum_id": index_to_datum_id[unique_matches[idx, 0]],
                    "ground_truth_id": index_to_groundtruth_id[
                        unique_matches[idx, 1]
                    ],
                    "prediction_id": index_to_prediction_id[
                        unique_matches[idx, 2]
                    ],
                }
            )

    return metric


def unpack_confusion_matrix_with_examples(
    metrics: dict[int, dict[int, Metric]],
    detailed_pairs: NDArray[np.float64],
    mask_tp: NDArray[np.bool_],
    mask_fp_fn_misclf: NDArray[np.bool_],
    mask_fp_unmatched: NDArray[np.bool_],
    mask_fn_unmatched: NDArray[np.bool_],
    index_to_datum_id: dict[int, str],
    index_to_groundtruth_id: dict[int, str],
    index_to_prediction_id: dict[int, str],
    index_to_label: dict[int, str],
) -> list[Metric]:

    ids = detailed_pairs[:, :5].astype(np.int32)

    mask_matched = mask_tp | mask_fp_fn_misclf

    return [
        _unpack_confusion_matrix_with_examples(
            metric=metric,
            ids=ids,
            mask_matched=mask_matched[iou_idx, score_idx],
            mask_fp_unmatched=mask_fp_unmatched[iou_idx, score_idx],
            mask_fn_unmatched=mask_fn_unmatched[iou_idx, score_idx],
            index_to_datum_id=index_to_datum_id,
            index_to_groundtruth_id=index_to_groundtruth_id,
            index_to_prediction_id=index_to_prediction_id,
            index_to_label=index_to_label,
        )
        for iou_idx, inner in metrics.items()
        for score_idx, metric in inner.items()
    ]



from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar

import numpy as np
from numpy.typing import NDArray
from shapely.geometry import Polygon as ShapelyPolygon


@dataclass
class BoundingBox:
    """
    Represents a bounding box with associated labels and optional scores.

    Parameters
    ----------
    uid : str
        A unique identifier.
    xmin : float
        The minimum x-coordinate of the bounding box.
    xmax : float
        The maximum x-coordinate of the bounding box.
    ymin : float
        The minimum y-coordinate of the bounding box.
    ymax : float
        The maximum y-coordinate of the bounding box.
    labels : list of str
        List of labels associated with the bounding box.
    scores : list of float, optional
        Confidence scores corresponding to each label. Defaults to an empty list.
    metadata : dict[str, Any], optional
        A dictionary containing any metadata to be used within filtering operations.

    Examples
    --------
    Ground Truth Example:

    >>> bbox = BoundingBox(uid="xyz", xmin=10.0, xmax=50.0, ymin=20.0, ymax=60.0, labels=['cat'])

    Prediction Example:

    >>> bbox = BoundingBox(
    ...     uid="abc",
    ...     xmin=10.0, xmax=50.0, ymin=20.0, ymax=60.0,
    ...     labels=['cat', 'dog'], scores=[0.9, 0.1]
    ... )
    """

    uid: str
    xmin: float
    xmax: float
    ymin: float
    ymax: float
    labels: list[str]
    scores: list[float] = field(default_factory=list)
    metadata: dict[str, Any] | None = None

    def __post_init__(self):
        if len(self.scores) == 0 and len(self.labels) != 1:
            raise ValueError(
                "Ground truths must be defined with no scores and a single label. If you meant to define a prediction, then please include one score for every label provided."
            )
        if len(self.scores) > 0 and len(self.labels) != len(self.scores):
            raise ValueError(
                "If scores are defined, there must be a 1:1 pairing with labels."
            )

    @property
    def extrema(self) -> tuple[float, float, float, float]:
        """
        Returns the annotation's data representation.

        Returns
        -------
        tuple[float, float, float, float]
            A tuple in the form (xmin, xmax, ymin, ymax).
        """
        return (self.xmin, self.xmax, self.ymin, self.ymax)


@dataclass
class Polygon:
    """
    Represents a polygon shape with associated labels and optional scores.

    Parameters
    ----------
    uid : str
        A unique identifier.
    shape : shapely.geometry.Polygon
        A Shapely polygon object representing the shape.
    labels : list of str
        List of labels associated with the polygon.
    scores : list of float, optional
        Confidence scores corresponding to each label. Defaults to an empty list.
    metadata : dict[str, Any], optional
        A dictionary containing any metadata to be used within filtering operations.

    Examples
    --------
    Ground Truth Example:

    >>> from shapely.geometry import Polygon as ShapelyPolygon
    >>> shape = ShapelyPolygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    >>> polygon = Polygon(uid="xyz", shape=shape, labels=['building'])

    Prediction Example:

    >>> polygon = Polygon(
    ...     uid="abc", shape=shape, labels=['building'], scores=[0.95]
    ... )
    """

    uid: str
    shape: ShapelyPolygon
    labels: list[str]
    scores: list[float] = field(default_factory=list)
    metadata: dict[str, Any] | None = None

    def __post_init__(self):
        if not isinstance(self.shape, ShapelyPolygon):
            raise TypeError("shape must be of type shapely.geometry.Polygon.")
        if self.shape.is_empty:
            raise ValueError("Polygon is empty.")

        if len(self.scores) == 0 and len(self.labels) != 1:
            raise ValueError(
                "Ground truths must be defined with no scores and a single label. If you meant to define a prediction, then please include one score for every label provided."
            )
        if len(self.scores) > 0 and len(self.labels) != len(self.scores):
            raise ValueError(
                "If scores are defined, there must be a 1:1 pairing with labels."
            )


@dataclass
class Bitmask:
    """
    Represents a binary mask with associated labels and optional scores.

    Parameters
    ----------
    uid : str
        A unique identifier.
    mask : NDArray[np.bool_]
        A NumPy array of boolean values representing the mask.
    labels : list of str
        List of labels associated with the mask.
    scores : list of float, optional
        Confidence scores corresponding to each label. Defaults to an empty list.
    metadata : dict[str, Any], optional
        A dictionary containing any metadata to be used within filtering operations.

    Examples
    --------
    Ground Truth Example:

    >>> import numpy as np
    >>> mask = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> bitmask = Bitmask(uid="abc", mask=mask, labels=['tree'])

    Prediction Example:

    >>> bitmask = Bitmask(
    ...     uid="xyz", mask=mask, labels=['tree'], scores=[0.85]
    ... )
    """

    uid: str
    mask: NDArray[np.bool_]
    labels: list[str]
    scores: list[float] = field(default_factory=list)
    metadata: dict[str, Any] | None = None

    def __post_init__(self):

        if (
            not isinstance(self.mask, np.ndarray)
            or self.mask.dtype != np.bool_
        ):
            raise ValueError(
                "Expected mask to be of type `NDArray[np.bool_]`."
            )
        elif not self.mask.any():
            raise ValueError("Mask does not define any object instances.")

        if len(self.scores) == 0 and len(self.labels) != 1:
            raise ValueError(
                "Ground truths must be defined with no scores and a single label. If you meant to define a prediction, then please include one score for every label provided."
            )
        if len(self.scores) > 0 and len(self.labels) != len(self.scores):
            raise ValueError(
                "If scores are defined, there must be a 1:1 pairing with labels."
            )


AnnotationType = TypeVar("AnnotationType", BoundingBox, Polygon, Bitmask)


@dataclass
class Detection(Generic[AnnotationType]):
    """
    Detection data structure holding ground truths and predictions for object detection tasks.

    Parameters
    ----------
    uid : str
        Unique identifier for the image or sample.
    groundtruths : list[BoundingBox] | list[Polygon] | list[Bitmask]
        List of ground truth annotations.
    predictions : list[BoundingBox] | list[Polygon] | list[Bitmask]
        List of predicted annotations.
    metadata : dict[str, Any], optional
        A dictionary containing any metadata to be used within filtering operations.

    Examples
    --------
    >>> bbox_gt = BoundingBox(xmin=10, xmax=50, ymin=20, ymax=60, labels=['cat'])
    >>> bbox_pred = BoundingBox(
    ...     xmin=12, xmax=48, ymin=22, ymax=58, labels=['cat'], scores=[0.9]
    ... )
    >>> detection = Detection(
    ...     uid='image_001',
    ...     groundtruths=[bbox_gt],
    ...     predictions=[bbox_pred]
    ... )
    """

    uid: str
    groundtruths: list[AnnotationType]
    predictions: list[AnnotationType]
    metadata: dict[str, Any] | None = None

    def __post_init__(self):
        for prediction in self.predictions:
            if len(prediction.scores) != len(prediction.labels):
                raise ValueError(
                    "Predictions must provide a score for every label."
                )



from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError


def validate_statements(
    response: dict[str, list[dict[str, str]]],
    key: str,
    allowed_values: set[str] | None = None,
    enforce_length: int | None = None,
):
    if key not in response:
        raise InvalidLLMResponseError(
            f"LLM did not include key '{key}' in its response: {response}"
        )
    elif (
        not isinstance(key, str)
        or not isinstance(response[key], list)
        or not all([isinstance(v, str) for v in response[key]])
    ):
        raise InvalidLLMResponseError(
            f"LLM response should follow the format 'dict[str, list[str]': {response}"
        )
    elif allowed_values is not None and not all(
        [v in allowed_values for v in response[key]]
    ):
        raise InvalidLLMResponseError(
            f"LLM response contains values from outside the allowed set {allowed_values}: {response}"
        )
    elif enforce_length is not None and enforce_length != len(response[key]):
        raise InvalidLLMResponseError(
            f"LLM response does not match input size of '{enforce_length}': {response}"
        )


def validate_verdicts(
    response: dict[str, list[dict[str, str]]],
    key: str,
    allowed_values: set[str] | None = None,
    enforce_length: int | None = None,
):
    if key not in response:
        raise InvalidLLMResponseError(
            f"LLM did not include key '{key}' in its response: {response}"
        )
    elif not isinstance(key, str) or not isinstance(response[key], list):
        raise InvalidLLMResponseError(
            f"LLM response should follow the format 'dict[str, list[dict[str, str]]]': {response}"
        )
    elif enforce_length is not None and enforce_length != len(response[key]):
        raise InvalidLLMResponseError(
            f"LLM response does not match input size of '{enforce_length}': {response}"
        )

    for value in response[key]:
        if not isinstance(value, dict):
            raise InvalidLLMResponseError(
                f"LLM response should follow the format 'dict[str, list[dict[str, str]]]': {response}"
            )
        elif set(value.keys()) != {"verdict", "analysis"}:
            raise InvalidLLMResponseError(
                f"LLM response is malformed. Inner dictionaries should only contain keys 'verdict' and 'analysis': {response} "
            )
        elif (
            allowed_values is not None
            and value["verdict"] not in allowed_values
        ):
            raise InvalidLLMResponseError(
                f"LLM response contains verdicts from outside the allowed set {allowed_values}: {response}"
            )



import warnings
from dataclasses import dataclass, field
from typing import Any

import numpy as np
from numpy.typing import NDArray


@dataclass
class Bitmask:
    """
    Represents a binary mask with an associated semantic label.

    Parameters
    ----------
    mask : NDArray[np.bool_]
        A NumPy array of boolean values representing the mask.
    label : str
        The semantic label associated with the mask.
    metadata : dict[str, Any], optional
        A dictionary containing any metadata to be used within filtering operations.

    Examples
    --------
    >>> import numpy as np
    >>> mask = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> bitmask = Bitmask(mask=mask, label='ocean')
    """

    mask: NDArray[np.bool_]
    label: str
    metadata: dict[str, Any] | None = None

    def __post_init__(self):
        if self.mask.dtype != np.bool_:
            raise ValueError(
                f"Bitmask recieved mask with dtype '{self.mask.dtype}'."
            )


@dataclass
class Segmentation:
    """
    Segmentation data structure holding ground truth and prediction bitmasks for semantic segmentation tasks.

    Parameters
    ----------
    uid : str
        Unique identifier for the image or sample.
    groundtruths : List[Bitmask]
        List of ground truth bitmasks.
    predictions : List[Bitmask]
        List of predicted bitmasks.
    shape : tuple of int, optional
        The shape of the segmentation masks. This is set automatically after initialization.
    size : int, optional
        The total number of pixels in the masks. This is set automatically after initialization.
    metadata : dict[str, Any], optional
        A dictionary containing any metadata to be used within filtering operations.

    Examples
    --------
    >>> import numpy as np
    >>> mask1 = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> groundtruth = Bitmask(mask=mask1, label='object')
    >>> mask2 = np.array([[False, True], [True, False]], dtype=np.bool_)
    >>> prediction = Bitmask(mask=mask2, label='object')
    >>> segmentation = Segmentation(
    ...     uid='123',
    ...     groundtruths=[groundtruth],
    ...     predictions=[prediction]
    ... )
    """

    uid: str
    groundtruths: list[Bitmask]
    predictions: list[Bitmask]
    shape: tuple[int, ...]
    size: int = field(default=0)
    metadata: dict[str, Any] | None = None

    def __post_init__(self):

        if len(self.shape) != 2 or self.shape[0] <= 0 or self.shape[1] <= 0:
            raise ValueError(
                f"segmentations must be 2-dimensional and have non-zero dimensions. Recieved shape '{self.shape}'"
            )
        self.size = self.shape[0] * self.shape[1]

        self._validate_bitmasks(self.groundtruths, "ground truth")
        self._validate_bitmasks(self.predictions, "prediction")

    def _validate_bitmasks(self, bitmasks: list[Bitmask], key: str):
        mask_accumulation = None
        mask_overlap_accumulation = None
        for idx, bitmask in enumerate(bitmasks):
            if not isinstance(bitmask, Bitmask):
                raise ValueError(f"expected 'Bitmask', got '{bitmask}'")
            if self.shape != bitmask.mask.shape:
                raise ValueError(
                    f"{key} masks for datum '{self.uid}' should have shape '{self.shape}'. Received mask with shape '{bitmask.mask.shape}'"
                )

            if mask_accumulation is None:
                mask_accumulation = bitmask.mask.copy()
                mask_overlap_accumulation = np.zeros_like(mask_accumulation)
            elif np.logical_and(mask_accumulation, bitmask.mask).any():
                mask_overlap = np.logical_and(mask_accumulation, bitmask.mask)
                bitmasks[idx].mask[mask_overlap] = False
                mask_overlap_accumulation = (
                    mask_overlap_accumulation | mask_overlap
                )
            else:
                mask_accumulation = mask_accumulation | bitmask.mask
        if (
            mask_overlap_accumulation is not None
            and mask_overlap_accumulation.any()
        ):
            count = mask_overlap_accumulation.sum()
            total = mask_overlap_accumulation.size
            warnings.warn(
                f"{key} masks for datum '{self.uid}' had {count} / {total} pixels overlapped."
            )



from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
from numpy.typing import NDArray

from valor_lite.cache import FileCacheReader, MemoryCacheReader


@dataclass
class EvaluatorInfo:
    number_of_datums: int = 0
    number_of_groundtruth_annotations: int = 0
    number_of_prediction_annotations: int = 0
    number_of_labels: int = 0
    number_of_rows: int = 0
    metadata_fields: list[tuple[str, str | pa.DataType]] | None = None


def generate_detailed_cache_path(path: str | Path) -> Path:
    return Path(path) / "detailed"


def generate_ranked_cache_path(path: str | Path) -> Path:
    return Path(path) / "ranked"


def generate_temporary_cache_path(path: str | Path) -> Path:
    return Path(path) / "tmp"


def generate_metadata_path(path: str | Path) -> Path:
    return Path(path) / "metadata.json"


def generate_detailed_schema(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> pa.Schema:
    metadata_fields = metadata_fields if metadata_fields else []
    reserved_fields = [
        ("datum_uid", pa.string()),
        ("datum_id", pa.int64()),
        # groundtruth
        ("gt_uid", pa.string()),
        ("gt_id", pa.int64()),
        ("gt_label", pa.string()),
        ("gt_label_id", pa.int64()),
        # prediction
        ("pd_uid", pa.string()),
        ("pd_id", pa.int64()),
        ("pd_label", pa.string()),
        ("pd_label_id", pa.int64()),
        ("pd_score", pa.float64()),
        # pair
        ("iou", pa.float64()),
    ]

    # validate
    reserved_field_names = {f[0] for f in reserved_fields}
    metadata_field_names = {f[0] for f in metadata_fields}
    if conflicting := reserved_field_names & metadata_field_names:
        raise ValueError(
            f"metadata fields {conflicting} conflict with reserved fields"
        )

    return pa.schema(reserved_fields + metadata_fields)


def generate_ranked_schema(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> pa.Schema:
    reserved_detailed_fields = [
        ("datum_uid", pa.string()),
        ("datum_id", pa.int64()),
        # groundtruth
        ("gt_id", pa.int64()),
        ("gt_label_id", pa.int64()),
        # prediction
        ("pd_id", pa.int64()),
        ("pd_label_id", pa.int64()),
        ("pd_score", pa.float64()),
        # pair
        ("iou", pa.float64()),
    ]
    reserved_ranking_fields = [
        ("iou_prev", pa.float64()),
    ]
    metadata_fields = metadata_fields if metadata_fields else []

    # validate
    reserved_field_names = {
        f[0] for f in reserved_detailed_fields + reserved_ranking_fields
    }
    metadata_field_names = {f[0] for f in metadata_fields}
    if conflicting := reserved_field_names & metadata_field_names:
        raise ValueError(
            f"metadata fields {conflicting} conflict with reserved fields"
        )

    return pa.schema(
        [
            *reserved_detailed_fields,
            *metadata_fields,
            *reserved_ranking_fields,
        ]
    )


def encode_metadata_fields(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> dict[str, str]:
    metadata_fields = metadata_fields if metadata_fields else []
    return {k: str(v) for k, v in metadata_fields}


def decode_metadata_fields(
    encoded_metadata_fields: dict[str, str]
) -> list[tuple[str, str]]:
    return [(k, v) for k, v in encoded_metadata_fields.items()]


def extract_labels(
    reader: MemoryCacheReader | FileCacheReader,
    index_to_label_override: dict[int, str] | None = None,
) -> dict[int, str]:
    if index_to_label_override is not None:
        return index_to_label_override

    index_to_label = {}
    for tbl in reader.iterate_tables(
        columns=[
            "gt_label_id",
            "gt_label",
            "pd_label_id",
            "pd_label",
        ]
    ):

        # get gt labels
        gt_label_ids = tbl["gt_label_id"].to_numpy()
        gt_label_ids, gt_indices = np.unique(gt_label_ids, return_index=True)
        gt_labels = tbl["gt_label"].take(gt_indices).to_pylist()
        gt_labels = dict(zip(gt_label_ids.astype(int).tolist(), gt_labels))
        gt_labels.pop(-1, None)
        index_to_label.update(gt_labels)

        # get pd labels
        pd_label_ids = tbl["pd_label_id"].to_numpy()
        pd_label_ids, pd_indices = np.unique(pd_label_ids, return_index=True)
        pd_labels = tbl["pd_label"].take(pd_indices).to_pylist()
        pd_labels = dict(zip(pd_label_ids.astype(int).tolist(), pd_labels))
        pd_labels.pop(-1, None)
        index_to_label.update(pd_labels)

    return index_to_label


def extract_counts(
    reader: MemoryCacheReader | FileCacheReader,
    datums: pc.Expression | None = None,
    groundtruths: pc.Expression | None = None,
    predictions: pc.Expression | None = None,
):
    n_dts, n_gts, n_pds = 0, 0, 0
    for tbl in reader.iterate_tables(filter=datums):
        # count datums
        n_dts += int(np.unique(tbl["datum_id"].to_numpy()).shape[0])

        # count groundtruths
        if groundtruths is not None:
            gts = tbl.filter(groundtruths)["gt_id"].to_numpy()
        else:
            gts = tbl["gt_id"].to_numpy()
        n_gts += int(np.unique(gts[gts >= 0]).shape[0])

        # count predictions
        if predictions is not None:
            pds = tbl.filter(predictions)["pd_id"].to_numpy()
        else:
            pds = tbl["pd_id"].to_numpy()
        n_pds += int(np.unique(pds[pds >= 0]).shape[0])

    return n_dts, n_gts, n_pds


def extract_groundtruth_count_per_label(
    reader: MemoryCacheReader | FileCacheReader,
    number_of_labels: int,
    datums: pc.Expression | None = None,
) -> NDArray[np.uint64]:
    gt_counts_per_lbl = np.zeros(number_of_labels, dtype=np.uint64)
    for gts in reader.iterate_arrays(
        numeric_columns=["gt_id", "gt_label_id"],
        filter=datums,
    ):
        # count gts per label
        unique_ann = np.unique(gts[gts[:, 0] >= 0], axis=0)
        unique_labels, label_counts = np.unique(
            unique_ann[:, 1], return_counts=True
        )
        for label_id, count in zip(unique_labels, label_counts):
            gt_counts_per_lbl[int(label_id)] += int(count)

    return gt_counts_per_lbl



from valor_lite.text_generation.llm.generation import (
    generate_answer_correctness_verdicts,
    generate_answer_relevance_verdicts,
    generate_bias_verdicts,
    generate_claims,
    generate_context_precision_verdicts,
    generate_context_recall_verdicts,
    generate_context_relevance_verdicts,
    generate_faithfulness_verdicts,
    generate_hallucination_verdicts,
    generate_opinions,
    generate_statements,
    generate_summary_coherence,
    generate_toxicity_verdicts,
)
from valor_lite.text_generation.llm.integrations import ClientWrapper


def calculate_answer_correctness(
    client: ClientWrapper,
    system_prompt: str,
    query: str,
    response: str,
    groundtruths: list[str],
) -> float:
    """
    Compute answer correctness. Answer correctness is computed as an f1 score obtained
    by comparing prediction statements to ground truth statements.

    If there are multiple ground truths, then the f1 score is computed for each ground
    truth and the maximum score is returned.

    This metric was adapted from RAGAS. We follow a similar prompting strategy and
    computation, however we do not do a weighted sum with an answer similarity score
    using embeddings.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    query : str
        The user query.
    response : str
        A generated response.
    groundtruths : list[str]
        A list of ground truth references.

    Returns
    -------
    float
        The answer correctness score between 0 and 1. Higher values indicate that the
        answer is more correct. A score of 1 indicates that all statements in the
        prediction are supported by the ground truth and all statements in the ground
        truth are present in the prediction.
    """
    prediction_statements = generate_statements(
        client=client,
        system_prompt=system_prompt,
        text=response,
    )
    f1_scores = [0.0]
    for groundtruth in groundtruths:
        groundtruth_statements = generate_statements(
            client=client,
            system_prompt=system_prompt,
            text=groundtruth,
        )
        verdicts = generate_answer_correctness_verdicts(
            client=client,
            system_prompt=system_prompt,
            query=query,
            groundtruth_statements=groundtruth_statements,
            prediction_statements=prediction_statements,
        )

        tp = len(verdicts["TP"])
        fp = len(verdicts["FP"])
        fn = len(verdicts["FN"])

        f1_scores.append(tp / (tp + 0.5 * (fp + fn)) if tp > 0 else 0)

    return max(f1_scores)


def calculate_answer_relevance(
    client: ClientWrapper,
    system_prompt: str,
    query: str,
    response: str,
) -> float:
    """
    Compute answer relevance, the proportion of the model response that is
    relevant to the query, for a single piece of text.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    query : str
        The user query.
    response : str
        A generated response.

    Returns
    -------
    float
        The answer relevance score between 0 and 1. A score of 1 indicates that all
        statements are relevant to the query.
    """
    statements = generate_statements(
        client=client,
        system_prompt=system_prompt,
        text=response,
    )
    verdicts = generate_answer_relevance_verdicts(
        client=client,
        system_prompt=system_prompt,
        query=query,
        statements=statements,
    )
    if len(verdicts) == 0:
        return 0.0

    return sum(verdict["verdict"] == "yes" for verdict in verdicts) / len(
        verdicts
    )


def calculate_bias(
    client: ClientWrapper,
    system_prompt: str,
    response: str,
) -> float:
    """
    Compute bias, the proportion of model opinions that are biased.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    response : str
        A generated response.

    Returns
    -------
    float
        The bias score between 0 and 1. A score of 1 indicates that all opinions in
        the text are biased.
    """

    opinions = generate_opinions(
        client=client,
        system_prompt=system_prompt,
        text=response,
    )
    if len(opinions) == 0:
        return 0.0

    verdicts = generate_bias_verdicts(
        client=client,
        system_prompt=system_prompt,
        opinions=opinions,
    )
    return sum(verdict["verdict"] == "yes" for verdict in verdicts) / len(
        verdicts
    )


def calculate_context_precision(
    client: ClientWrapper,
    system_prompt: str,
    query: str,
    predicted_context: list[str],
    groundtruth_context: list[str],
) -> float:
    """
    Compute context precision, a score for evaluating the retrieval
    mechanism of a RAG model.

    First, an LLM is prompted to determine if each context in the context
    list is useful for producing the ground truth answer to the query.

    If there are multiple ground truths, then the verdict is "yes" for a
    context if that context is useful for producing any of the ground truth
    answers, and "no" otherwise.

    Then, using these verdicts, the context precision score is computed as
    a weighted sum of the precision at k for each k from 1 to the length
    of the context list.

    Note that the earlier a piece of context appears in the context list,
    the more important it is in the computation of this score. For example,
    the first context in the context list will be included in every precision
    at k computation, so will have a large influence on the final score,
    whereas the last context will only be used for the last precision at
    k computation, so will have a small influence on the final score.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    query : str
        The user query.
    response : str
        A generated response.
    predicted_context : list[str]
        A list of predicted context.
    groundtruths : list[str]
        A list of ground truth references.

    Returns
    -------
    float
        The context precision score between 0 and 1. A higher score indicates
        better context precision.
    """
    if len(predicted_context) == 0 and len(groundtruth_context) == 0:
        return 1.0
    elif len(predicted_context) == 0 or len(groundtruth_context) == 0:
        return 0.0

    # Get verdicts for each ground truth, and aggregate by setting the verdict for
    # a context to "yes" if the verdict is "yes" for any ground truth.
    aggregate_verdicts = ["no"] * len(predicted_context)
    for groundtruth in groundtruth_context:
        verdicts = generate_context_precision_verdicts(
            client=client,
            system_prompt=system_prompt,
            query=query,
            ordered_context_list=predicted_context,
            groundtruth=groundtruth,
        )
        for i in range(len(verdicts)):
            if verdicts[i]["verdict"] == "yes":
                aggregate_verdicts[i] = "yes"

    # Use the aggregate verdicts to compute the precision at k for each k.
    precision_at_k_list = []
    for k in range(1, len(predicted_context) + 1):
        # Only compute the precision at k if the kth context is relevant.
        if aggregate_verdicts[k - 1] == "yes":
            precision_at_k = (
                sum(verdict == "yes" for verdict in aggregate_verdicts[:k]) / k
            )
            precision_at_k_list.append(precision_at_k)

    # If none of the context are relevant, then the context precision is 0.
    if len(precision_at_k_list) == 0:
        return 0.0

    # Average over all the precision at k for which the kth context is relevant.
    return sum(precision_at_k_list) / len(precision_at_k_list)


def calculate_context_recall(
    client: ClientWrapper,
    system_prompt: str,
    predicted_context: list[str],
    groundtruth_context: list[str],
) -> float:
    """
    Compute context recall, a score for evaluating the retrieval mechanism of a RAG model.

    The context recall score is the proportion of statements in the ground truth
    that are attributable to the context list.

    If multiple ground truths are provided, then the context recall score is
    computed for each ground truth and the maximum score is returned.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    predicted_context : list[str]
        A list of predicted context.
    groundtruths : list[str]
        A list of ground truth references.

    Returns
    -------
    float
        The context recall score between 0 and 1. A score of 1 indicates that
        all ground truth statements are attributable to the contexts in the context list.
    """
    if len(predicted_context) == 0 and len(groundtruth_context) == 0:
        return 1.0
    elif len(predicted_context) == 0 or len(groundtruth_context) == 0:
        return 0.0

    scores = []
    for groundtruth in groundtruth_context:
        groundtruth_statements = generate_statements(
            client=client,
            system_prompt=system_prompt,
            text=groundtruth,
        )
        verdicts = generate_context_recall_verdicts(
            client=client,
            system_prompt=system_prompt,
            context_list=predicted_context,
            groundtruth_statements=groundtruth_statements,
        )
        scores.append(
            sum(verdict["verdict"] == "yes" for verdict in verdicts)
            / len(verdicts)
        )

    return max(scores)


def calculate_context_relevance(
    client: ClientWrapper,
    system_prompt: str,
    query: str,
    context: list[str],
) -> float:
    """
    Compute context relevance, the proportion of contexts in the context list
    that are relevant to the query.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    query : str
        The user query.
    context : list[str]
        A list of predicted context.

    Returns
    -------
    float
        The context relevance score between 0 and 1. A score of 0 indicates
        that none of the contexts are relevant and a score of 1 indicates
        that all of the contexts are relevant.
    """
    if len(context) == 0:
        return 0.0
    verdicts = generate_context_relevance_verdicts(
        client=client,
        system_prompt=system_prompt,
        query=query,
        context_list=context,
    )
    return sum(verdict["verdict"] == "yes" for verdict in verdicts) / len(
        verdicts
    )


def calculate_faithfulness(
    client: ClientWrapper,
    system_prompt: str,
    response: str,
    context: list[str],
) -> float:
    """
    Compute the faithfulness score. The faithfulness score is the proportion
    of claims in the text that are implied by the list of contexts. Claims
    that contradict the list of contexts and claims that are unrelated to
    the list of contexts both count against the score.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    response : str
        A generated response.
    context : list[str]
        A list of predicted context.

    Returns
    -------
    float
        The faithfulness score between 0 and 1. A score of 1 indicates that
        all claims in the text are implied by the list of contexts.
    """
    if len(context) == 0:
        return 0.0

    claims = generate_claims(
        client=client, system_prompt=system_prompt, text=response
    )

    # If there aren't any claims, then the text is perfectly faithful, as the text does not contain any non-faithful claims.
    if len(claims) == 0:
        return 1.0

    faithfulness_verdicts = generate_faithfulness_verdicts(
        client=client,
        system_prompt=system_prompt,
        claims=claims,
        context_list=context,
    )
    return sum(
        verdict["verdict"] == "yes" for verdict in faithfulness_verdicts
    ) / len(faithfulness_verdicts)


def calculate_hallucination(
    client: ClientWrapper,
    system_prompt: str,
    response: str,
    context: list[str],
) -> float:
    """
    Compute the hallucination score, the proportion of contexts in the context
    list that are contradicted by the text.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    response : str
        A generated response.
    context : list[str]
        A list of predicted context.

    Returns
    -------
    float
        The hallucination score between 0 and 1. A score of 1 indicates that
        all contexts are contradicted by the text.
    """
    if len(context) == 0:
        raise ValueError("Hallucination requires context to be calculated.")

    verdicts = generate_hallucination_verdicts(
        client=client,
        system_prompt=system_prompt,
        text=response,
        context_list=context,
    )
    return sum(verdict["verdict"] == "yes" for verdict in verdicts) / len(
        verdicts
    )


def calculate_summary_coherence(
    client: ClientWrapper,
    system_prompt: str,
    text: str,
    summary: str,
) -> int:
    """
    Compute summary coherence, the collective quality of a summary.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    text : str
        The original text.
    summary : str
        The generated summary.

    Returns
    -------
    int
        The summary coherence score between 1 and 5. A score of 1 indicates
        the lowest summary coherence and a score of 5 indicates the highest
        summary coherence.
    """
    return generate_summary_coherence(
        client=client,
        system_prompt=system_prompt,
        text=text,
        summary=summary,
    )


def calculate_toxicity(
    client: ClientWrapper,
    system_prompt: str,
    response: str,
) -> float:
    """
    Compute toxicity, the proportion of opinions that are toxic.

    Parameters
    ----------
    client : ClientWrapper
        The LLM client used to perform evaluation.
    system_prompt : str
        A system prompt to pass with the evaluation query.
    response : str
        A generated response.

    Returns
    -------
    Metric
        The toxicity score will be evaluated as a float between 0 and 1, with
        1 indicating that all opinions in the text are toxic.
    """
    opinions = generate_opinions(
        client=client,
        system_prompt=system_prompt,
        text=response,
    )
    if len(opinions) == 0:
        return 0.0

    verdicts = generate_toxicity_verdicts(
        client=client,
        system_prompt=system_prompt,
        opinions=opinions,
    )
    return sum(verdict["verdict"] == "yes" for verdict in verdicts) / len(
        verdicts
    )


def calculate_rouge_scores(
    prediction: str,
    references: str | list[str],
    rouge_types: list[str],
    use_stemmer: bool = False,
) -> dict[str, float]:
    """
    Calculate ROUGE scores for a prediction given some set of references.

    Parameters
    ----------
    prediction : str
        A generated response to score. Each prediction should be a string with tokens separated by spaces.
    references : str | list[str]
        A list of references for a given response. Each reference should be a string with tokens separated by spaces.
    rouge_types : list[str]
        A list of rouge types to calculate.
    use_stemmer: bool, default=False
        If True, uses Porter stemmer to strip word suffixes. Defaults to False.
    """
    import evaluate

    rouge = evaluate.load("rouge")

    metrics = rouge.compute(
        predictions=[prediction],
        references=[references],
        rouge_types=rouge_types,
        use_stemmer=use_stemmer,
        use_aggregator=False,  # aggregation gives us an average across all predictions, which isn't what we want
    )

    # find the max value for each prediction
    results = dict()
    if metrics is not None:
        for type_ in rouge_types:
            results[type_] = max(metrics[type_][0], 0.0)
    return results


def calculate_sentence_bleu(
    prediction: str,
    references: list[str],
    weights: tuple[float, ...] | list[float],
) -> float:
    """
    Calculate sentence BLEU scores for a of prediction - ground truth pair.

    Parameters
    ----------
    prediction : str
        A generated response to score. Each prediction should be a string with tokens separated by spaces.
    references : list[str]
        A list of references for a given response. Each reference should be a string with tokens separated by spaces.
    weights : tuple[float]
        The default BLEU calculates a score for up to 4-grams using uniform
        weights (this is called BLEU-4). To evaluate your translations with
        higher/lower order ngrams, use customized weights. Example: when accounting
        for up to 5-grams with uniform weights (this is called BLEU-5) use [1/5]*5
    """
    from nltk.tokenize import RegexpTokenizer
    from nltk.translate import bleu_score

    if len(weights) == 0:
        raise ValueError("At least one weight should be defined.")

    tokenizer = RegexpTokenizer(
        r"\w+|\$[\d]+|[^\s\.]+"
    )  # regex tokenizer that ignores periods

    tokenized_prediction = tokenizer.tokenize(prediction)
    tokenized_references = [tokenizer.tokenize(ref) for ref in references]

    # find the max value for each prediction
    result = float(
        bleu_score.sentence_bleu(
            references=tokenized_references,
            hypothesis=tokenized_prediction,
            weights=weights,
        ),  # type: ignore
    )
    return result if result >= 1e-9 else 0.0



import numpy as np
from numpy.typing import NDArray


def compute_intermediates(
    groundtruths: NDArray[np.bool_],
    predictions: NDArray[np.bool_],
    groundtruth_labels: NDArray[np.int64],
    prediction_labels: NDArray[np.int64],
    n_labels: int,
) -> NDArray[np.uint64]:
    """
    Computes an intermediate confusion matrix containing label counts.

    Parameters
    ----------
    groundtruths : NDArray[np.bool_]
        A 2-D array containing flattened bitmasks for each label.
    predictions : NDArray[np.bool_]
        A 2-D array containing flattened bitmasks for each label.
    groundtruth_labels : NDArray[np.int64]
        A 1-D array containing ground truth label indices.
    prediction_labels : NDArray[np.int64]
        A 1-D array containing prediction label indices.
    n_labels : int
        The number of unique labels.

    Returns
    -------
    NDArray[np.uint64]
        A 2-D confusion matrix with shape (n_labels + 1, n_labels + 1).
    """

    groundtruth_counts = groundtruths.sum(axis=1)
    prediction_counts = predictions.sum(axis=1)

    background_counts = np.logical_not(
        groundtruths.any(axis=0) | predictions.any(axis=0)
    ).sum()

    intersection_counts = np.logical_and(
        groundtruths[:, None, :],
        predictions[None, :, :],
    ).sum(axis=2)
    intersected_groundtruth_counts = intersection_counts.sum(axis=1)
    intersected_prediction_counts = intersection_counts.sum(axis=0)

    confusion_matrix = np.zeros((n_labels + 1, n_labels + 1), dtype=np.uint64)
    confusion_matrix[0, 0] = background_counts
    confusion_matrix[
        np.ix_(groundtruth_labels + 1, prediction_labels + 1)
    ] = intersection_counts
    confusion_matrix[0, prediction_labels + 1] = (
        prediction_counts - intersected_prediction_counts
    )
    confusion_matrix[groundtruth_labels + 1, 0] = (
        groundtruth_counts - intersected_groundtruth_counts
    )
    return confusion_matrix


def compute_metrics(
    confusion_matrix: NDArray[np.uint64],
) -> tuple[
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    float,
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
]:
    """
    Computes semantic segmentation metrics.

    Parameters
    ----------
    counts : NDArray[np.uint64]
        A 2-D confusion matrix with shape (n_labels + 1, n_labels + 1).

    Returns
    -------
    NDArray[np.float64]
        Precision.
    NDArray[np.float64]
        Recall.
    NDArray[np.float64]
        F1 Score.
    float
        Accuracy
    NDArray[np.float64]
        Confusion matrix containing IOU values.
    NDArray[np.float64]
        Unmatched prediction ratios.
    NDArray[np.float64]
        Unmatched ground truth ratios.
    """
    n_labels = confusion_matrix.shape[0] - 1
    n_pixels = confusion_matrix.sum()
    gt_counts = confusion_matrix[1:, :].sum(axis=1)
    pd_counts = confusion_matrix[:, 1:].sum(axis=0)

    # compute iou, unmatched_ground_truth and unmatched predictions
    intersection_ = confusion_matrix[1:, 1:]
    union_ = (
        gt_counts[:, np.newaxis] + pd_counts[np.newaxis, :] - intersection_
    )

    ious = np.zeros((n_labels, n_labels), dtype=np.float64)
    np.divide(
        intersection_,
        union_,
        where=union_ > 1e-9,
        out=ious,
    )

    unmatched_prediction_ratio = np.zeros((n_labels), dtype=np.float64)
    np.divide(
        confusion_matrix[0, 1:],
        pd_counts,
        where=pd_counts > 1e-9,
        out=unmatched_prediction_ratio,
    )

    unmatched_ground_truth_ratio = np.zeros((n_labels), dtype=np.float64)
    np.divide(
        confusion_matrix[1:, 0],
        gt_counts,
        where=gt_counts > 1e-9,
        out=unmatched_ground_truth_ratio,
    )

    # compute precision, recall, f1
    tp_counts = confusion_matrix.diagonal()[1:]

    precision = np.zeros(n_labels, dtype=np.float64)
    np.divide(tp_counts, pd_counts, where=pd_counts > 1e-9, out=precision)

    recall = np.zeros_like(precision)
    np.divide(tp_counts, gt_counts, where=gt_counts > 1e-9, out=recall)

    f1_score = np.zeros_like(precision)
    np.divide(
        2 * (precision * recall),
        (precision + recall),
        where=(precision + recall) > 0,
        out=f1_score,
    )

    # compute accuracy
    tp_count = confusion_matrix[1:, 1:].diagonal().sum()
    background_count = confusion_matrix[0, 0]
    accuracy = (
        (tp_count + background_count) / n_pixels if n_pixels > 0 else 0.0
    )

    return (
        precision,
        recall,
        f1_score,
        accuracy,
        ious,
        unmatched_prediction_ratio,
        unmatched_ground_truth_ratio,
    )



from collections import defaultdict

import numpy as np
import pyarrow as pa
from numpy.typing import NDArray

from valor_lite.classification.metric import Metric, MetricType


def unpack_precision_recall(
    counts: NDArray[np.uint64],
    precision: NDArray[np.float64],
    recall: NDArray[np.float64],
    accuracy: NDArray[np.float64],
    f1_score: NDArray[np.float64],
    score_thresholds: list[float],
    hardmax: bool,
    index_to_label: dict[int, str],
) -> dict[MetricType, list[Metric]]:

    metrics = defaultdict(list)

    metrics[MetricType.Accuracy] = [
        Metric.accuracy(
            value=float(accuracy[score_idx]),
            score_threshold=score_threshold,
            hardmax=hardmax,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    for label_idx, label in index_to_label.items():
        for score_idx, score_threshold in enumerate(score_thresholds):
            kwargs = {
                "label": label,
                "hardmax": hardmax,
                "score_threshold": score_threshold,
            }
            row = counts[:, label_idx]
            metrics[MetricType.Counts].append(
                Metric.counts(
                    tp=int(row[score_idx, 0]),
                    fp=int(row[score_idx, 1]),
                    fn=int(row[score_idx, 2]),
                    tn=int(row[score_idx, 3]),
                    **kwargs,
                )
            )

            metrics[MetricType.Precision].append(
                Metric.precision(
                    value=float(precision[score_idx, label_idx]),
                    **kwargs,
                )
            )
            metrics[MetricType.Recall].append(
                Metric.recall(
                    value=float(recall[score_idx, label_idx]),
                    **kwargs,
                )
            )
            metrics[MetricType.F1].append(
                Metric.f1_score(
                    value=float(f1_score[score_idx, label_idx]),
                    **kwargs,
                )
            )
    return metrics


def unpack_rocauc(
    rocauc: NDArray[np.float64],
    mean_rocauc: float,
    index_to_label: dict[int, str],
) -> dict[MetricType, list[Metric]]:
    metrics = {}
    metrics[MetricType.ROCAUC] = [
        Metric.roc_auc(
            value=float(rocauc[label_idx]),
            label=label,
        )
        for label_idx, label in index_to_label.items()
    ]
    metrics[MetricType.mROCAUC] = [
        Metric.mean_roc_auc(
            value=float(mean_rocauc),
        )
    ]
    return metrics


def unpack_confusion_matrix(
    confusion_matrices: NDArray[np.uint64],
    unmatched_groundtruths: NDArray[np.uint64],
    index_to_label: dict[int, str],
    score_thresholds: list[float],
    hardmax: bool,
) -> list[Metric]:
    metrics = []
    for score_idx, score_thresh in enumerate(score_thresholds):
        cm_dict = {}
        ugt_dict = {}
        for idx, label in index_to_label.items():
            ugt_dict[label] = int(unmatched_groundtruths[score_idx, idx])
            for pidx, plabel in index_to_label.items():
                if label not in cm_dict:
                    cm_dict[label] = {}
                cm_dict[label][plabel] = int(
                    confusion_matrices[score_idx, idx, pidx]
                )
        metrics.append(
            Metric.confusion_matrix(
                confusion_matrix=cm_dict,
                unmatched_ground_truths=ugt_dict,
                score_threshold=score_thresh,
                hardmax=hardmax,
            )
        )
    return metrics


def create_mapping(
    tbl: pa.Table,
    pairs: NDArray[np.float64],
    index: int,
    id_col: str,
    uid_col: str,
) -> dict[int, str]:
    col = pairs[:, index].astype(np.int64)
    values, indices = np.unique(col, return_index=True)
    indices = indices[values >= 0]
    return {
        tbl[id_col][idx].as_py(): tbl[uid_col][idx].as_py() for idx in indices
    }


def unpack_examples(
    ids: NDArray[np.int64],
    mask_tp: NDArray[np.bool_],
    mask_fn: NDArray[np.bool_],
    mask_fp: NDArray[np.bool_],
    score_thresholds: list[float],
    hardmax: bool,
    index_to_datum_id: dict[int, str],
    index_to_label: dict[int, str],
) -> list[Metric]:
    metrics = []
    unique_datums = np.unique(ids[:, 0])
    for datum_index in unique_datums:
        mask_datum = ids[:, 0] == datum_index
        mask_datum_tp = mask_tp & mask_datum
        mask_datum_fp = mask_fp & mask_datum
        mask_datum_fn = mask_fn & mask_datum

        datum_id = index_to_datum_id[datum_index]
        for score_idx, score_thresh in enumerate(score_thresholds):

            unique_tp = np.unique(
                # extract true-positive (datum_id, gt_id, pd_id) pairs
                ids[np.ix_(mask_datum_tp[score_idx], (0, 1, 2))],
                axis=0,
            )
            unique_fp = np.unique(
                # extract false-positive (datum_id, pd_id) pairs
                ids[np.ix_(mask_datum_fp[score_idx], (0, 2))],
                axis=0,
            )
            unique_fn = np.unique(
                # extract false-negative (datum_id, gt_id)
                ids[np.ix_(mask_datum_fn[score_idx], (0, 1))],
                axis=0,
            )

            tp = [index_to_label[row[1]] for row in unique_tp]
            fp = [
                index_to_label[row[1]]
                for row in unique_fp
                if index_to_label[row[1]] not in tp
            ]
            fn = [
                index_to_label[row[1]]
                for row in unique_fn
                if index_to_label[row[1]] not in tp
            ]
            metrics.append(
                Metric.examples(
                    datum_id=datum_id,
                    true_positives=tp,
                    false_negatives=fn,
                    false_positives=fp,
                    score_threshold=score_thresh,
                    hardmax=hardmax,
                )
            )
    return metrics


def create_empty_confusion_matrix_with_examples(
    score_threshold: float,
    hardmax: bool,
    index_to_label: dict[int, str],
) -> Metric:
    unmatched_groundtruths = dict()
    confusion_matrix = dict()
    for label in index_to_label.values():
        unmatched_groundtruths[label] = {"count": 0, "examples": []}
        confusion_matrix[label] = {}
        for plabel in index_to_label.values():
            confusion_matrix[label][plabel] = {"count": 0, "examples": []}

    return Metric.confusion_matrix_with_examples(
        confusion_matrix=confusion_matrix,
        unmatched_ground_truths=unmatched_groundtruths,
        score_threshold=score_threshold,
        hardmax=hardmax,
    )


def _unpack_confusion_matrix_with_examples(
    metric: Metric,
    ids: NDArray[np.int64],
    scores: NDArray[np.float64],
    winners: NDArray[np.bool_],
    mask_matched: NDArray[np.bool_],
    mask_unmatched_fn: NDArray[np.bool_],
    index_to_datum_id: dict[int, str],
    index_to_label: dict[int, str],
):
    if not isinstance(metric.value, dict):
        raise TypeError("expected metric to contain a dictionary value")

    mask_valid_gts = ids[:, 1] >= 0
    mask_valid_pds = ids[:, 2] >= 0

    valid_matches = ids[mask_valid_gts & mask_valid_pds]
    valid_gts = ids[mask_valid_gts]

    n_matched = 0
    unique_matches = np.empty((1, 3))
    if valid_matches.size > 0:
        unique_matches, unique_match_indices = np.unique(
            # extract matched (datum_id, gt_id, pd_id) pairs
            valid_matches[np.ix_(mask_matched, (0, 1, 2))],  # type: ignore[reportArgumentType]
            axis=0,
            return_index=True,
        )
        scores = scores[mask_matched][unique_match_indices]
        n_matched = unique_matches.shape[0]

    n_unmatched_groundtruths = 0
    unique_unmatched_groundtruths = np.empty((1, 2))
    if valid_gts.size > 0:
        unique_unmatched_groundtruths = np.unique(
            # extract unmatched false-negative (datum_id, gt_id) pairs
            valid_gts[np.ix_(mask_unmatched_fn, (0, 1))],  # type: ignore[reportArgumentType]
            axis=0,
        )
        unique_unmatched_groundtruths = unique_unmatched_groundtruths[
            unique_unmatched_groundtruths[:, 1] >= 0
        ]
        n_unmatched_groundtruths = unique_unmatched_groundtruths.shape[0]

    n_max = max(n_matched, n_unmatched_groundtruths)
    for idx in range(n_max):
        if idx < n_matched:
            datum_id = index_to_datum_id[unique_matches[idx, 0]]
            glabel = index_to_label[unique_matches[idx, 1]]
            plabel = index_to_label[unique_matches[idx, 2]]
            score = float(scores[idx])

            metric.value["confusion_matrix"][glabel][plabel]["count"] += 1
            metric.value["confusion_matrix"][glabel][plabel][
                "examples"
            ].append(
                {
                    "datum_id": datum_id,
                    "score": score,
                }
            )
        if idx < n_unmatched_groundtruths:
            datum_id = index_to_datum_id[unique_unmatched_groundtruths[idx, 0]]
            label = index_to_label[unique_unmatched_groundtruths[idx, 1]]

            metric.value["unmatched_ground_truths"][label]["count"] += 1
            metric.value["unmatched_ground_truths"][label]["examples"].append(
                {"datum_id": datum_id}
            )

    return metric


def unpack_confusion_matrix_with_examples(
    metrics: dict[int, Metric],
    ids: NDArray[np.int64],
    scores: NDArray[np.float64],
    winners: NDArray[np.bool_],
    mask_matched: NDArray[np.bool_],
    mask_unmatched_fn: NDArray[np.bool_],
    index_to_datum_id: dict[int, str],
    index_to_label: dict[int, str],
) -> list[Metric]:
    return [
        _unpack_confusion_matrix_with_examples(
            metric,
            ids=ids,
            scores=scores,
            winners=winners,
            mask_matched=mask_matched[score_idx, :],
            mask_unmatched_fn=mask_unmatched_fn[score_idx, :],
            index_to_datum_id=index_to_datum_id,
            index_to_label=index_to_label,
        )
        for score_idx, metric in metrics.items()
    ]



class EmptyEvaluatorError(Exception):
    def __init__(self):
        super().__init__(
            "evaluator cannot be finalized as it contains no data"
        )


class EmptyCacheError(Exception):
    def __init__(self):
        super().__init__("cache contains no data")


class EmptyFilterError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class InternalCacheError(Exception):
    def __init__(self, message: str):
        super().__init__(message)



from enum import IntFlag, auto

import numpy as np
import pyarrow as pa
import shapely
from numpy.typing import NDArray

EPSILON = 1e-9


def compute_bbox_iou(data: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for axis-aligned bounding boxes.

    Takes data with shape (N, 8):

    Index 0 - xmin for Box 1
    Index 1 - xmax for Box 1
    Index 2 - ymin for Box 1
    Index 3 - ymax for Box 1
    Index 4 - xmin for Box 2
    Index 5 - xmax for Box 2
    Index 6 - ymin for Box 2
    Index 7 - ymax for Box 2

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of bounding box pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """
    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]

    xmin1, xmax1, ymin1, ymax1 = (
        data[:, 0, 0],
        data[:, 0, 1],
        data[:, 0, 2],
        data[:, 0, 3],
    )
    xmin2, xmax2, ymin2, ymax2 = (
        data[:, 1, 0],
        data[:, 1, 1],
        data[:, 1, 2],
        data[:, 1, 3],
    )

    xmin = np.maximum(xmin1, xmin2)
    ymin = np.maximum(ymin1, ymin2)
    xmax = np.minimum(xmax1, xmax2)
    ymax = np.minimum(ymax1, ymax2)

    intersection_width = np.maximum(0, xmax - xmin)
    intersection_height = np.maximum(0, ymax - ymin)
    intersection_area = intersection_width * intersection_height

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    union_area = area1 + area2 - intersection_area

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_area,
        union_area,
        where=union_area >= EPSILON,
        out=ious,
    )
    return ious


def compute_bitmask_iou(data: NDArray[np.bool_]) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for bitmasks.

    Takes data with shape (N, 2):

    Index 0 - first bitmask
    Index 1 - second bitmask

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of bitmask pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """

    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]
    lhs = data[:, 0, :, :].reshape(n_pairs, -1)
    rhs = data[:, 1, :, :].reshape(n_pairs, -1)

    lhs_sum = lhs.sum(axis=1)
    rhs_sum = rhs.sum(axis=1)

    intersection_ = np.logical_and(lhs, rhs).sum(axis=1)
    union_ = lhs_sum + rhs_sum - intersection_

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_,
        union_,
        where=union_ >= EPSILON,
        out=ious,
    )
    return ious


def compute_polygon_iou(
    data: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for shapely polygons.

    Takes data with shape (N, 2):

    Index 0 - first polygon
    Index 1 - second polygon

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of polygon pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """

    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]

    lhs = data[:, 0]
    rhs = data[:, 1]

    intersections = shapely.intersection(lhs, rhs)
    intersection_areas = shapely.area(intersections)

    unions = shapely.union(lhs, rhs)
    union_areas = shapely.area(unions)

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_areas,
        union_areas,
        where=union_areas >= EPSILON,
        out=ious,
    )
    return ious


def rank_pairs(
    sorted_pairs: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.intp]]:
    """
    Prunes and ranks prediction pairs.

    Should result in a single pair per prediction annotation.

    Parameters
    ----------
    sorted_pairs : NDArray[np.float64]
        Ranked annotation pairs.
        Index 0 - Datum Index
        Index 1 - GroundTruth Index
        Index 2 - Prediction Index
        Index 3 - GroundTruth Label Index
        Index 4 - Prediction Label Index
        Index 5 - IOU
        Index 6 - Score

    Returns
    -------
    NDArray[float64]
        Ranked prediction pairs.
    NDArray[intp]
        Indices of ranked prediction pairs.
    """

    # remove unmatched ground truths
    mask_predictions = sorted_pairs[:, 2] >= 0.0
    pairs = sorted_pairs[mask_predictions]
    indices = np.where(mask_predictions)[0]

    # find best fits for prediction
    mask_label_match = np.isclose(pairs[:, 3], pairs[:, 4])
    matched_predictions = np.unique(pairs[mask_label_match, 2])

    mask_unmatched_predictions = ~np.isin(pairs[:, 2], matched_predictions)

    pairs = pairs[mask_label_match | mask_unmatched_predictions]
    indices = indices[mask_label_match | mask_unmatched_predictions]

    # only keep the highest ranked prediction (datum_id, prediction_id, predicted_label_id)
    _, unique_indices = np.unique(
        pairs[:, [0, 2, 4]], axis=0, return_index=True
    )
    pairs = pairs[unique_indices]
    indices = indices[unique_indices]

    # np.unique orders its results by value, we need to sort the indices to maintain the results of the lexsort
    sorted_indices = np.lexsort(
        (
            -pairs[:, 5],  # iou
            -pairs[:, 6],  # score
        )
    )
    pairs = pairs[sorted_indices]
    indices = indices[sorted_indices]

    return pairs, indices


def calculate_ranking_boundaries(
    ranked_pairs: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Determine IOU boundaries for computing AP across chunks.

    Parameters
    ----------
    ranked_pairs : NDArray[np.float64]
        Ranked annotation pairs.
        Index 0 - Datum Index
        Index 1 - GroundTruth Index
        Index 2 - Prediction Index
        Index 3 - GroundTruth Label Index
        Index 4 - Prediction Label Index
        Index 5 - IOU
        Index 6 - Score

    Returns
    -------
    NDArray[np.float64]
        A 1-D array containing the lower IOU boundary for classifying pairs as true-positive across chunks.
    """
    ids = ranked_pairs[:, (0, 1, 2, 3, 4)].astype(np.int64)
    gts = ids[:, (0, 1, 3)]
    gt_labels = ids[:, 3]
    pd_labels = ids[:, 4]
    ious = ranked_pairs[:, 5]

    # set default boundary to 2.0 as it will be used to check lower boundary in range [0-1].
    iou_boundary = np.ones_like(ious) * 2

    mask_matching_labels = gt_labels == pd_labels
    mask_valid_gts = gts[:, 1] >= 0
    unique_gts = np.unique(gts[mask_valid_gts], axis=0)
    for gt in unique_gts:
        mask_gt = (gts == gt).all(axis=1)
        mask_gt &= mask_matching_labels
        if mask_gt.sum() <= 1:
            iou_boundary[mask_gt] = 0.0
            continue

        running_max = np.maximum.accumulate(ious[mask_gt])
        mask_rmax = np.isclose(running_max, ious[mask_gt])
        mask_rmax[1:] &= running_max[1:] > running_max[:-1]
        mask_gt[mask_gt] &= mask_rmax

        indices = np.where(mask_gt)[0]

        iou_boundary[indices[0]] = 0.0
        iou_boundary[indices[1:]] = ious[indices[:-1]]

    return iou_boundary


def rank_table(tbl: pa.Table) -> pa.Table:
    """Rank table for AP computation."""
    numeric_columns = [
        "datum_id",
        "gt_id",
        "pd_id",
        "gt_label_id",
        "pd_label_id",
        "iou",
        "pd_score",
    ]
    sorting_args = [
        ("pd_score", "descending"),
        ("iou", "descending"),
    ]

    # initial sort
    sorted_tbl = tbl.sort_by(sorting_args)
    pairs = np.column_stack(
        [sorted_tbl[col].to_numpy() for col in numeric_columns]
    )

    # rank pairs
    ranked_pairs, indices = rank_pairs(pairs)
    ranked_tbl = sorted_tbl.take(indices)

    # find boundaries
    lower_iou_bound = calculate_ranking_boundaries(ranked_pairs)
    ranked_tbl = ranked_tbl.append_column(
        pa.field("iou_prev", pa.float64()),
        pa.array(lower_iou_bound, type=pa.float64()),
    )

    return ranked_tbl


def compute_counts(
    ranked_pairs: NDArray[np.float64],
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
    number_of_groundtruths_per_label: NDArray[np.uint64],
    number_of_labels: int,
    running_counts: NDArray[np.uint64],
    pr_curve: NDArray[np.float64],
) -> NDArray[np.uint64]:
    """
    Computes Object Detection metrics.

    Precision-recall curve and running counts are updated in-place.

    Parameters
    ----------
    ranked_pairs : NDArray[np.float64]
        A ranked array summarizing the IOU calculations of one or more pairs.
        Index 0 - Datum Index
        Index 1 - GroundTruth Index
        Index 2 - Prediction Index
        Index 3 - GroundTruth Label Index
        Index 4 - Prediction Label Index
        Index 5 - IOU
        Index 6 - Score
        Index 7 - IOU Lower Boundary
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.
    number_of_groundtruths_per_label : NDArray[np.uint64]
        A 1-D array containing total number of ground truths per label.
    number_of_labels : int
        Total number of unique labels.
    running_counts : NDArray[np.uint64]
        A 2-D array containing running counts of total predictions and true-positive. This array is mutated.
    pr_curve : NDArray[np.float64]
        A 2-D array containing 101-point binning of precision and score over a fixed recall interval. This array is mutated.

    Returns
    -------
    NDArray[uint64]
        Batched counts of TP, FP, FN.
    """
    n_rows = ranked_pairs.shape[0]
    n_labels = number_of_labels
    n_ious = iou_thresholds.shape[0]
    n_scores = score_thresholds.shape[0]

    # initialize result arrays
    counts = np.zeros((n_ious, n_scores, 3, n_labels), dtype=np.uint64)

    # start computation
    ids = ranked_pairs[:, :5].astype(np.int64)
    gt_ids = ids[:, 1]
    gt_labels = ids[:, 3]
    pd_labels = ids[:, 4]
    ious = ranked_pairs[:, 5]
    scores = ranked_pairs[:, 6]
    prev_ious = ranked_pairs[:, 7]

    unique_pd_labels, _ = np.unique(pd_labels, return_index=True)

    running_total_count = np.zeros(
        (n_ious, n_rows),
        dtype=np.uint64,
    )
    running_tp_count = np.zeros_like(running_total_count)
    running_gt_count = number_of_groundtruths_per_label[pd_labels]

    mask_score_nonzero = scores > EPSILON
    mask_gt_exists = gt_ids >= 0.0
    mask_labels_match = np.isclose(gt_labels, pd_labels)

    mask_gt_exists_labels_match = mask_gt_exists & mask_labels_match

    mask_tp = mask_score_nonzero & mask_gt_exists_labels_match
    mask_fp = mask_score_nonzero

    for iou_idx in range(n_ious):
        mask_iou_curr = ious >= iou_thresholds[iou_idx]
        mask_iou_prev = prev_ious < iou_thresholds[iou_idx]
        mask_iou = mask_iou_curr & mask_iou_prev

        mask_tp_outer = mask_tp & mask_iou
        mask_fp_outer = mask_fp & (
            (~mask_gt_exists_labels_match & mask_iou) | ~mask_iou
        )

        for score_idx in range(n_scores):
            mask_score_thresh = scores >= score_thresholds[score_idx]

            mask_tp_inner = mask_tp_outer & mask_score_thresh
            mask_fp_inner = mask_fp_outer & mask_score_thresh

            # create true-positive mask score threshold
            tp_candidates = ids[mask_tp_inner]
            _, indices_gt_unique = np.unique(
                tp_candidates[:, [0, 1, 3]], axis=0, return_index=True
            )
            mask_gt_unique = np.zeros(tp_candidates.shape[0], dtype=np.bool_)
            mask_gt_unique[indices_gt_unique] = True

            true_positives_mask = np.zeros(n_rows, dtype=np.bool_)
            true_positives_mask[mask_tp_inner] = mask_gt_unique

            mask_fp_inner |= mask_tp_inner & ~true_positives_mask

            # calculate intermediates
            counts[iou_idx, score_idx, 0, :] = np.bincount(
                pd_labels,
                weights=true_positives_mask,
                minlength=n_labels,
            )
            # fp count
            counts[iou_idx, score_idx, 1, :] = np.bincount(
                pd_labels[mask_fp_inner],
                minlength=n_labels,
            )

        # count running tp and total for AP
        for pd_label in unique_pd_labels:
            mask_pd_label = pd_labels == pd_label
            total_count = mask_pd_label.sum()
            if total_count == 0:
                continue

            # running total prediction count
            running_total_count[iou_idx, mask_pd_label] = np.arange(
                running_counts[iou_idx, pd_label, 0] + 1,
                running_counts[iou_idx, pd_label, 0] + total_count + 1,
            )
            running_counts[iou_idx, pd_label, 0] += total_count

            # running true-positive count
            mask_tp_for_counting = mask_pd_label & mask_tp_outer
            tp_count = mask_tp_for_counting.sum()
            running_tp_count[iou_idx, mask_tp_for_counting] = np.arange(
                running_counts[iou_idx, pd_label, 1] + 1,
                running_counts[iou_idx, pd_label, 1] + tp_count + 1,
            )
            running_counts[iou_idx, pd_label, 1] += tp_count

    # calculate running precision-recall points for AP
    precision = np.zeros_like(running_total_count, dtype=np.float64)
    np.divide(
        running_tp_count,
        running_total_count,
        where=running_total_count > 0,
        out=precision,
    )
    recall = np.zeros_like(running_total_count, dtype=np.float64)
    np.divide(
        running_tp_count,
        running_gt_count,
        where=running_gt_count > 0,
        out=recall,
    )
    recall_index = np.floor(recall * 100.0).astype(np.int32)

    # sort precision in descending order
    precision_indices = np.argsort(-precision, axis=1)

    # populate precision-recall curve
    for iou_idx in range(n_ious):
        labeled_recall = np.hstack(
            [
                pd_labels.reshape(-1, 1),
                recall_index[iou_idx, :].reshape(-1, 1),
            ]
        )

        # extract maximum score per (label, recall) bin
        # arrays are already ordered by descending score
        lr_pairs, recall_indices = np.unique(
            labeled_recall, return_index=True, axis=0
        )
        li = lr_pairs[:, 0]
        ri = lr_pairs[:, 1]
        pr_curve[iou_idx, li, ri, 1] = np.maximum(
            pr_curve[iou_idx, li, ri, 1],
            scores[recall_indices],
        )

        # extract maximum precision per (label, recall) bin
        # reorder arrays into descending precision order
        indices = precision_indices[iou_idx]
        sorted_precision = precision[iou_idx, indices]
        sorted_labeled_recall = labeled_recall[indices]
        lr_pairs, recall_indices = np.unique(
            sorted_labeled_recall, return_index=True, axis=0
        )
        li = lr_pairs[:, 0]
        ri = lr_pairs[:, 1]
        pr_curve[iou_idx, li, ri, 0] = np.maximum(
            pr_curve[iou_idx, li, ri, 0],
            sorted_precision[recall_indices],
        )

    return counts


def compute_precision_recall_f1(
    counts: NDArray[np.uint64],
    number_of_groundtruths_per_label: NDArray[np.uint64],
) -> NDArray[np.float64]:

    prec_rec_f1 = np.zeros_like(counts, dtype=np.float64)

    # alias
    tp_count = counts[:, :, 0, :]
    fp_count = counts[:, :, 1, :]
    tp_fp_count = tp_count + fp_count

    # calculate component metrics
    np.divide(
        tp_count,
        tp_fp_count,
        where=tp_fp_count > 0,
        out=prec_rec_f1[:, :, 0, :],
    )
    np.divide(
        tp_count,
        number_of_groundtruths_per_label,
        where=number_of_groundtruths_per_label > 0,
        out=prec_rec_f1[:, :, 1, :],
    )
    p = prec_rec_f1[:, :, 0, :]
    r = prec_rec_f1[:, :, 1, :]
    np.divide(
        2 * np.multiply(p, r),
        (p + r),
        where=(p + r) > EPSILON,
        out=prec_rec_f1[:, :, 2, :],
    )
    return prec_rec_f1


def compute_average_recall(prec_rec_f1: NDArray[np.float64]):
    recall = prec_rec_f1[:, :, 1, :]
    average_recall = recall.mean(axis=0)
    mAR = average_recall.mean(axis=-1)
    return average_recall, mAR


def compute_average_precision(pr_curve: NDArray[np.float64]):
    n_ious = pr_curve.shape[0]
    n_labels = pr_curve.shape[1]

    # initialize result arrays
    average_precision = np.zeros((n_ious, n_labels), dtype=np.float64)
    mAP = np.zeros(n_ious, dtype=np.float64)

    # calculate average precision
    running_max_precision = np.zeros((n_ious, n_labels), dtype=np.float64)
    running_max_score = np.zeros((n_labels), dtype=np.float64)
    for recall in range(100, -1, -1):

        # running max precision
        running_max_precision = np.maximum(
            pr_curve[:, :, recall, 0],
            running_max_precision,
        )
        pr_curve[:, :, recall, 0] = running_max_precision

        # running max score
        running_max_score = np.maximum(
            pr_curve[:, :, recall, 1],
            running_max_score,
        )
        pr_curve[:, :, recall, 1] = running_max_score

        average_precision += running_max_precision

    average_precision = average_precision / 101.0

    # calculate mAP and mAR
    if average_precision.size > 0:
        mAP = average_precision.mean(axis=1)

    return average_precision, mAP, pr_curve


def _isin(
    data: NDArray,
    subset: NDArray,
) -> NDArray[np.bool_]:
    """
    Creates a mask of rows that exist within the subset.

    Parameters
    ----------
    data : NDArray[np.int32]
        An array with shape (N, 2).
    subset : NDArray[np.int32]
        An array with shape (M, 2) where N >= M.

    Returns
    -------
    NDArray[np.bool_]
        Returns a bool mask with shape (N,).
    """
    combined_data = (data[:, 0].astype(np.int64) << 32) | data[:, 1].astype(
        np.int32
    )
    combined_subset = (subset[:, 0].astype(np.int64) << 32) | subset[
        :, 1
    ].astype(np.int32)
    mask = np.isin(combined_data, combined_subset, assume_unique=False)
    return mask


class PairClassification(IntFlag):
    NULL = auto()
    TP = auto()
    FP_FN_MISCLF = auto()
    FP_UNMATCHED = auto()
    FN_UNMATCHED = auto()


def mask_pairs_greedily(
    pairs: NDArray[np.float64],
):
    groundtruths = pairs[:, 1].astype(np.int32)
    predictions = pairs[:, 2].astype(np.int32)

    # Pre‑allocate "seen" flags for every possible x and y
    max_gt = groundtruths.max()
    max_pd = predictions.max()
    used_gt = np.zeros(max_gt + 1, dtype=np.bool_)
    used_pd = np.zeros(max_pd + 1, dtype=np.bool_)

    # This mask will mark which pairs to keep
    keep = np.zeros(pairs.shape[0], dtype=bool)

    for idx in range(groundtruths.shape[0]):
        gidx = groundtruths[idx]
        pidx = predictions[idx]

        if not (gidx < 0 or pidx < 0 or used_gt[gidx] or used_pd[pidx]):
            keep[idx] = True
            used_gt[gidx] = True
            used_pd[pidx] = True

    mask_matches = _isin(
        data=pairs[:, (1, 2)],
        subset=np.unique(pairs[np.ix_(keep, (1, 2))], axis=0),  # type: ignore - np.ix_ typing
    )

    return mask_matches


def compute_pair_classifications(
    detailed_pairs: NDArray[np.float64],
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
) -> tuple[
    NDArray[np.bool_], NDArray[np.bool_], NDArray[np.bool_], NDArray[np.bool_]
]:
    """
    Compute detailed counts.

    Takes data with shape (N, 7):

    Index 0 - Datum Index
    Index 1 - GroundTruth Index
    Index 2 - Prediction Index
    Index 3 - GroundTruth Label Index
    Index 4 - Prediction Label Index
    Index 5 - IOU
    Index 6 - Score

    Parameters
    ----------
    detailed_pairs : NDArray[np.float64]
        An unsorted array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.

    Returns
    -------
    NDArray[np.uint8]
        Confusion matrix.
    """
    n_pairs = detailed_pairs.shape[0]
    n_ious = iou_thresholds.shape[0]
    n_scores = score_thresholds.shape[0]

    pair_classifications = np.zeros(
        (n_ious, n_scores, n_pairs),
        dtype=np.uint8,
    )

    ids = detailed_pairs[:, :5].astype(np.int32)
    groundtruths = ids[:, (0, 1)]
    predictions = ids[:, (0, 2)]
    gt_ids = ids[:, 1]
    pd_ids = ids[:, 2]
    gt_labels = ids[:, 3]
    pd_labels = ids[:, 4]
    ious = detailed_pairs[:, 5]
    scores = detailed_pairs[:, 6]

    mask_gt_exists = gt_ids > -0.5
    mask_pd_exists = pd_ids > -0.5
    mask_label_match = np.isclose(gt_labels, pd_labels)
    mask_score_nonzero = scores > EPSILON
    mask_iou_nonzero = ious > EPSILON

    mask_gt_pd_exists = mask_gt_exists & mask_pd_exists
    mask_gt_pd_match = mask_gt_pd_exists & mask_label_match

    mask_matched_pairs = mask_pairs_greedily(pairs=detailed_pairs)

    for iou_idx in range(n_ious):
        mask_iou_threshold = ious >= iou_thresholds[iou_idx]
        mask_iou = mask_iou_nonzero & mask_iou_threshold
        for score_idx in range(n_scores):
            mask_score_threshold = scores >= score_thresholds[score_idx]
            mask_score = mask_score_nonzero & mask_score_threshold

            mask_thresholded_matched_pairs = (
                mask_matched_pairs & mask_iou & mask_score
            )

            mask_true_positives = (
                mask_thresholded_matched_pairs & mask_gt_pd_match
            )
            mask_misclf = mask_thresholded_matched_pairs & ~mask_gt_pd_match

            mask_groundtruths_in_thresholded_matched_pairs = _isin(
                data=groundtruths,
                subset=np.unique(
                    groundtruths[mask_thresholded_matched_pairs], axis=0
                ),
            )
            mask_predictions_in_thresholded_matched_pairs = _isin(
                data=predictions,
                subset=np.unique(
                    predictions[mask_thresholded_matched_pairs], axis=0
                ),
            )

            mask_unmatched_predictions = (
                ~mask_predictions_in_thresholded_matched_pairs
                & mask_pd_exists
                & mask_score
            )
            mask_unmatched_groundtruths = (
                ~mask_groundtruths_in_thresholded_matched_pairs
                & mask_gt_exists
            )

            # classify pairings
            pair_classifications[
                iou_idx, score_idx, mask_true_positives
            ] |= np.uint8(PairClassification.TP)
            pair_classifications[iou_idx, score_idx, mask_misclf] |= np.uint8(
                PairClassification.FP_FN_MISCLF
            )
            pair_classifications[
                iou_idx, score_idx, mask_unmatched_predictions
            ] |= np.uint8(PairClassification.FP_UNMATCHED)
            pair_classifications[
                iou_idx, score_idx, mask_unmatched_groundtruths
            ] |= np.uint8(PairClassification.FN_UNMATCHED)

    mask_tp = np.bitwise_and(pair_classifications, PairClassification.TP) > 0
    mask_fp_fn_misclf = (
        np.bitwise_and(pair_classifications, PairClassification.FP_FN_MISCLF)
        > 0
    )
    mask_fp_unmatched = (
        np.bitwise_and(pair_classifications, PairClassification.FP_UNMATCHED)
        > 0
    )
    mask_fn_unmatched = (
        np.bitwise_and(pair_classifications, PairClassification.FN_UNMATCHED)
        > 0
    )

    return (
        mask_tp,
        mask_fp_fn_misclf,
        mask_fp_unmatched,
        mask_fn_unmatched,
    )


def compute_confusion_matrix(
    detailed_pairs: NDArray[np.float64],
    mask_tp: NDArray[np.bool_],
    mask_fp_fn_misclf: NDArray[np.bool_],
    mask_fp_unmatched: NDArray[np.bool_],
    mask_fn_unmatched: NDArray[np.bool_],
    number_of_labels: int,
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
):
    n_ious = iou_thresholds.size
    n_scores = score_thresholds.size
    ids = detailed_pairs[:, :5].astype(np.int64)

    # initialize arrays
    confusion_matrices = np.zeros(
        (n_ious, n_scores, number_of_labels, number_of_labels), dtype=np.uint64
    )
    unmatched_groundtruths = np.zeros(
        (n_ious, n_scores, number_of_labels), dtype=np.uint64
    )
    unmatched_predictions = np.zeros_like(unmatched_groundtruths)

    mask_matched = mask_tp | mask_fp_fn_misclf
    for iou_idx in range(n_ious):
        for score_idx in range(n_scores):
            # matched annotations
            unique_pairs = np.unique(
                ids[np.ix_(mask_matched[iou_idx, score_idx], (0, 1, 2, 3, 4))],  # type: ignore - numpy ix_ typing
                axis=0,
            )
            unique_labels, unique_label_counts = np.unique(
                unique_pairs[:, (3, 4)], axis=0, return_counts=True
            )
            confusion_matrices[
                iou_idx, score_idx, unique_labels[:, 0], unique_labels[:, 1]
            ] = unique_label_counts

            # unmatched groundtruths
            unique_pairs = np.unique(
                ids[np.ix_(mask_fn_unmatched[iou_idx, score_idx], (0, 1, 3))],  # type: ignore - numpy ix_ typing
                axis=0,
            )
            unique_labels, unique_label_counts = np.unique(
                unique_pairs[:, 2], return_counts=True
            )
            unmatched_groundtruths[
                iou_idx, score_idx, unique_labels
            ] = unique_label_counts

            # unmatched predictions
            unique_pairs = np.unique(
                ids[np.ix_(mask_fp_unmatched[iou_idx, score_idx], (0, 2, 4))],  # type: ignore - numpy ix_ typing
                axis=0,
            )
            unique_labels, unique_label_counts = np.unique(
                unique_pairs[:, 2], return_counts=True
            )
            unmatched_predictions[
                iou_idx, score_idx, unique_labels
            ] = unique_label_counts

    return confusion_matrices, unmatched_groundtruths, unmatched_predictions



from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
from numpy.typing import NDArray

from valor_lite.cache import FileCacheReader, MemoryCacheReader


@dataclass
class EvaluatorInfo:
    number_of_rows: int = 0
    number_of_datums: int = 0
    number_of_labels: int = 0
    metadata_fields: list[tuple[str, str]] | None = None


def generate_cache_path(path: str | Path) -> Path:
    return Path(path) / "cache"


def generate_intermediate_cache_path(path: str | Path) -> Path:
    return Path(path) / "intermediate"


def generate_roc_curve_cache_path(path: str | Path) -> Path:
    return Path(path) / "roc_curve"


def generate_metadata_path(path: str | Path) -> Path:
    return Path(path) / "metadata.json"


def generate_schema(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> pa.Schema:
    metadata_fields = metadata_fields if metadata_fields else []
    reserved_fields = [
        ("datum_uid", pa.string()),
        ("datum_id", pa.int64()),
        # groundtruth
        ("gt_label", pa.string()),
        ("gt_label_id", pa.int64()),
        # prediction
        ("pd_label", pa.string()),
        ("pd_label_id", pa.int64()),
        ("pd_score", pa.float64()),
        ("pd_winner", pa.bool_()),
        # pair
        ("match", pa.bool_()),
    ]

    # validate
    reserved_field_names = {f[0] for f in reserved_fields}
    metadata_field_names = {f[0] for f in metadata_fields}
    if conflicting := reserved_field_names & metadata_field_names:
        raise ValueError(
            f"metadata fields {conflicting} conflict with reserved fields"
        )
    return pa.schema(reserved_fields + metadata_fields)


def generate_intermediate_schema() -> pa.Schema:
    return pa.schema(
        [
            ("pd_label_id", pa.int64()),
            ("pd_score", pa.float64()),
            ("match", pa.bool_()),
        ]
    )


def generate_roc_curve_schema() -> pa.Schema:
    return pa.schema(
        [
            ("pd_label_id", pa.int64()),
            ("cumulative_fp", pa.uint64()),
            ("cumulative_tp", pa.uint64()),
        ]
    )


def encode_metadata_fields(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> dict[str, str]:
    metadata_fields = metadata_fields if metadata_fields else []
    return {k: str(v) for k, v in metadata_fields}


def decode_metadata_fields(
    encoded_metadata_fields: dict[str, str]
) -> list[tuple[str, str]]:
    return [(k, v) for k, v in encoded_metadata_fields.items()]


def extract_labels(
    reader: MemoryCacheReader | FileCacheReader,
    index_to_label_override: dict[int, str] | None = None,
) -> dict[int, str]:
    if index_to_label_override is not None:
        return index_to_label_override

    index_to_label = {}
    for tbl in reader.iterate_tables(
        columns=[
            "gt_label_id",
            "gt_label",
            "pd_label_id",
            "pd_label",
        ]
    ):

        # get gt labels
        gt_label_ids = tbl["gt_label_id"].to_numpy()
        gt_label_ids, gt_indices = np.unique(gt_label_ids, return_index=True)
        gt_labels = tbl["gt_label"].take(gt_indices).to_pylist()
        gt_labels = dict(zip(gt_label_ids.astype(int).tolist(), gt_labels))
        gt_labels.pop(-1, None)
        index_to_label.update(gt_labels)

        # get pd labels
        pd_label_ids = tbl["pd_label_id"].to_numpy()
        pd_label_ids, pd_indices = np.unique(pd_label_ids, return_index=True)
        pd_labels = tbl["pd_label"].take(pd_indices).to_pylist()
        pd_labels = dict(zip(pd_label_ids.astype(int).tolist(), pd_labels))
        pd_labels.pop(-1, None)
        index_to_label.update(pd_labels)

    return index_to_label


def extract_counts(
    reader: MemoryCacheReader | FileCacheReader,
    datums: pc.Expression | None = None,
):
    n_dts = 0
    for tbl in reader.iterate_tables(filter=datums):
        n_dts += int(np.unique(tbl["datum_id"].to_numpy()).shape[0])
    return n_dts


def extract_groundtruth_count_per_label(
    reader: MemoryCacheReader | FileCacheReader,
    number_of_labels: int,
    datums: pc.Expression | None = None,
    groundtruths: pc.Expression | None = None,
    predictions: pc.Expression | None = None,
) -> NDArray[np.uint64]:

    # count ground truth and prediction label occurences
    label_counts = np.zeros((number_of_labels, 2), dtype=np.uint64)
    for tbl in reader.iterate_tables(filter=datums):

        # count unique gt labels
        gt_expr = pc.field("gt_label_id") >= 0
        if groundtruths is not None:
            gt_expr &= groundtruths
        gt_tbl = tbl.filter(gt_expr)
        gt_ids = np.column_stack(
            [gt_tbl[col].to_numpy() for col in ["datum_id", "gt_label_id"]]
        ).astype(np.int64)
        unique_gts = np.unique(gt_ids, axis=0)
        unique_gt_labels, gt_label_counts = np.unique(
            unique_gts[:, 1], return_counts=True
        )
        label_counts[unique_gt_labels, 0] += gt_label_counts.astype(np.uint64)

        # count unique pd labels
        pd_expr = pc.field("pd_label_id") >= 0
        if predictions is not None:
            pd_expr &= predictions
        pd_tbl = tbl.filter(pd_expr)
        pd_ids = np.column_stack(
            [pd_tbl[col].to_numpy() for col in ["datum_id", "pd_label_id"]]
        ).astype(np.int64)
        unique_pds = np.unique(pd_ids, axis=0)
        unique_pd_labels, pd_label_counts = np.unique(
            unique_pds[:, 1], return_counts=True
        )
        label_counts[unique_pd_labels, 1] += pd_label_counts.astype(np.uint64)

    return label_counts



import json
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from valor_lite.cache import (
    FileCacheReader,
    FileCacheWriter,
    MemoryCacheReader,
    MemoryCacheWriter,
    compute,
)
from valor_lite.exceptions import EmptyCacheError
from valor_lite.object_detection.computation import (
    compute_average_precision,
    compute_average_recall,
    compute_confusion_matrix,
    compute_counts,
    compute_pair_classifications,
    compute_precision_recall_f1,
    rank_table,
)
from valor_lite.object_detection.metric import Metric, MetricType
from valor_lite.object_detection.shared import (
    EvaluatorInfo,
    decode_metadata_fields,
    encode_metadata_fields,
    extract_counts,
    extract_groundtruth_count_per_label,
    extract_labels,
    generate_detailed_cache_path,
    generate_detailed_schema,
    generate_metadata_path,
    generate_ranked_cache_path,
    generate_ranked_schema,
)
from valor_lite.object_detection.utilities import (
    create_empty_confusion_matrix_with_examples,
    create_mapping,
    unpack_confusion_matrix,
    unpack_confusion_matrix_with_examples,
    unpack_examples,
    unpack_precision_recall_into_metric_lists,
)


class Builder:
    def __init__(
        self,
        detailed_writer: MemoryCacheWriter | FileCacheWriter,
        ranked_writer: MemoryCacheWriter | FileCacheWriter,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        self._detailed_writer = detailed_writer
        self._ranked_writer = ranked_writer
        self._metadata_fields = metadata_fields

    @classmethod
    def in_memory(
        cls,
        batch_size: int = 10_000,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        """
        Create an in-memory evaluator cache.

        Parameters
        ----------
        batch_size : int, default=10_000
            The target number of rows to buffer before writing to the cache. Defaults to 10_000.
        metadata_fields : list[tuple[str, str | pa.DataType]], optional
            Optional datum metadata field definitions.
        """
        # create cache
        detailed_writer = MemoryCacheWriter.create(
            schema=generate_detailed_schema(metadata_fields),
            batch_size=batch_size,
        )
        ranked_writer = MemoryCacheWriter.create(
            schema=generate_ranked_schema(metadata_fields),
            batch_size=batch_size,
        )

        return cls(
            detailed_writer=detailed_writer,
            ranked_writer=ranked_writer,
            metadata_fields=metadata_fields,
        )

    @classmethod
    def persistent(
        cls,
        path: str | Path,
        batch_size: int = 10_000,
        rows_per_file: int = 100_000,
        compression: str = "snappy",
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        """
        Create a persistent file-based evaluator cache.

        Parameters
        ----------
        path : str | Path
            Where to store the file-based cache.
        batch_size : int, default=10_000
            The target number of rows to buffer before writing to the cache. Defaults to 10_000.
        rows_per_file : int, default=100_000
            The target number of rows to store per cache file. Defaults to 100_000.
        compression : str, default="snappy"
            The compression methods used when writing cache files.
        metadata_fields : list[tuple[str, str | pa.DataType]], optional
            Optional metadata field definitions.
        """
        path = Path(path)

        # create caches
        detailed_writer = FileCacheWriter.create(
            path=generate_detailed_cache_path(path),
            schema=generate_detailed_schema(metadata_fields),
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )
        ranked_writer = FileCacheWriter.create(
            path=generate_ranked_cache_path(path),
            schema=generate_ranked_schema(metadata_fields),
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )

        # write metadata
        metadata_path = generate_metadata_path(path)
        with open(metadata_path, "w") as f:
            encoded_types = encode_metadata_fields(metadata_fields)
            json.dump(encoded_types, f, indent=2)

        return cls(
            detailed_writer=detailed_writer,
            ranked_writer=ranked_writer,
            metadata_fields=metadata_fields,
        )

    def _rank(self, batch_size: int = 1_000):
        """Perform pair ranking over the detailed cache."""

        detailed_reader = self._detailed_writer.to_reader()
        compute.sort(
            source=detailed_reader,
            sink=self._ranked_writer,
            batch_size=batch_size,
            sorting=[
                ("pd_score", "descending"),
                ("iou", "descending"),
            ],
            columns=[
                field.name
                for field in self._ranked_writer.schema
                if field.name != "iou_prev"
            ],
            table_sort_override=rank_table,
        )
        self._ranked_writer.flush()

    def finalize(
        self,
        batch_size: int = 1_000,
        index_to_label_override: dict[int, str] | None = None,
    ):
        """
        Performs data finalization and preprocessing.

        Parameters
        ----------
        batch_size : int, default=1_000
            Sets the batch size for reading. Defaults to 1_000.
        index_to_label_override : dict[int, str], optional
            Pre-configures label mapping. Used when operating over filtered subsets.
        """
        self._detailed_writer.flush()
        if self._detailed_writer.count_rows() == 0:
            raise EmptyCacheError()

        self._detailed_writer.sort_by(
            [
                ("pd_score", "descending"),
                ("iou", "descending"),
            ]
        )
        detailed_reader = self._detailed_writer.to_reader()

        # extract labels
        index_to_label = extract_labels(
            reader=detailed_reader,
            index_to_label_override=index_to_label_override,
        )

        # populate ranked cache
        self._rank(batch_size)

        ranked_reader = self._ranked_writer.to_reader()
        return Evaluator(
            detailed_reader=detailed_reader,
            ranked_reader=ranked_reader,
            index_to_label=index_to_label,
            metadata_fields=self._metadata_fields,
        )


class Evaluator:
    def __init__(
        self,
        detailed_reader: MemoryCacheReader | FileCacheReader,
        ranked_reader: MemoryCacheReader | FileCacheReader,
        index_to_label: dict[int, str],
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        self._detailed_reader = detailed_reader
        self._ranked_reader = ranked_reader
        self._index_to_label = index_to_label
        self._metadata_fields = metadata_fields

    @property
    def info(self) -> EvaluatorInfo:
        return self.get_info()

    def get_info(
        self,
        datums: pc.Expression | None = None,
        groundtruths: pc.Expression | None = None,
        predictions: pc.Expression | None = None,
    ) -> EvaluatorInfo:
        info = EvaluatorInfo()
        info.number_of_rows = self._detailed_reader.count_rows()
        info.metadata_fields = self._metadata_fields
        info.number_of_labels = len(self._index_to_label)
        (
            info.number_of_datums,
            info.number_of_groundtruth_annotations,
            info.number_of_prediction_annotations,
        ) = extract_counts(
            reader=self._detailed_reader,
            datums=datums,
            groundtruths=groundtruths,
            predictions=predictions,
        )
        return info

    @classmethod
    def load(
        cls,
        path: str | Path,
        index_to_label_override: dict[int, str] | None = None,
    ):
        # validate path
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Directory does not exist: {path}")
        elif not path.is_dir():
            raise NotADirectoryError(
                f"Path exists but is not a directory: {path}"
            )

        detailed_reader = FileCacheReader.load(
            generate_detailed_cache_path(path)
        )
        ranked_reader = FileCacheReader.load(generate_ranked_cache_path(path))

        # extract labels from cache
        index_to_label = extract_labels(
            reader=detailed_reader,
            index_to_label_override=index_to_label_override,
        )

        # read config
        metadata_path = generate_metadata_path(path)
        metadata_fields = None
        with open(metadata_path, "r") as f:
            encoded_metadata_types = json.load(f)
            metadata_fields = decode_metadata_fields(encoded_metadata_types)

        return cls(
            detailed_reader=detailed_reader,
            ranked_reader=ranked_reader,
            index_to_label=index_to_label,
            metadata_fields=metadata_fields,
        )

    def filter(
        self,
        datums: pc.Expression | None = None,
        groundtruths: pc.Expression | None = None,
        predictions: pc.Expression | None = None,
        batch_size: int = 1_000,
        path: str | Path | None = None,
    ) -> "Evaluator":
        """
        Filter evaluator cache.

        Parameters
        ----------
        datums : pc.Expression | None = None
            A filter expression used to filter datums.
        groundtruths : pc.Expression | None = None
            A filter expression used to filter ground truth annotations.
        predictions : pc.Expression | None = None
            A filter expression used to filter predictions.
        batch_size : int
            The maximum number of rows read into memory per file.
        path : str | Path
            Where to store the filtered cache if storing on disk.

        Returns
        -------
        Evaluator
            A new evaluator object containing the filtered cache.
        """
        from valor_lite.object_detection.loader import Loader

        if isinstance(self._detailed_reader, FileCacheReader):
            if not path:
                raise ValueError(
                    "expected path to be defined for file-based loader"
                )
            loader = Loader.persistent(
                path=path,
                batch_size=self._detailed_reader.batch_size,
                rows_per_file=self._detailed_reader.rows_per_file,
                compression=self._detailed_reader.compression,
                metadata_fields=self._metadata_fields,
            )
        else:
            loader = Loader.in_memory(
                batch_size=self._detailed_reader.batch_size,
                metadata_fields=self._metadata_fields,
            )

        for tbl in self._detailed_reader.iterate_tables(filter=datums):
            columns = (
                "datum_id",
                "gt_id",
                "pd_id",
                "gt_label_id",
                "pd_label_id",
                "iou",
                "pd_score",
            )
            pairs = np.column_stack([tbl[col].to_numpy() for col in columns])

            n_pairs = pairs.shape[0]
            gt_ids = pairs[:, (0, 1)].astype(np.int64)
            pd_ids = pairs[:, (0, 2)].astype(np.int64)

            if groundtruths is not None:
                mask_valid_gt = np.zeros(n_pairs, dtype=np.bool_)
                gt_tbl = tbl.filter(groundtruths)
                gt_pairs = np.column_stack(
                    [gt_tbl[col].to_numpy() for col in ("datum_id", "gt_id")]
                ).astype(np.int64)
                for gt in np.unique(gt_pairs, axis=0):
                    mask_valid_gt |= (gt_ids == gt).all(axis=1)
            else:
                mask_valid_gt = np.ones(n_pairs, dtype=np.bool_)

            if predictions is not None:
                mask_valid_pd = np.zeros(n_pairs, dtype=np.bool_)
                pd_tbl = tbl.filter(predictions)
                pd_pairs = np.column_stack(
                    [pd_tbl[col].to_numpy() for col in ("datum_id", "pd_id")]
                ).astype(np.int64)
                for pd in np.unique(pd_pairs, axis=0):
                    mask_valid_pd |= (pd_ids == pd).all(axis=1)
            else:
                mask_valid_pd = np.ones(n_pairs, dtype=np.bool_)

            mask_valid = mask_valid_gt | mask_valid_pd
            mask_valid_gt &= mask_valid
            mask_valid_pd &= mask_valid

            # filter out invalid gt_id, gt_label_id by setting to -1.0
            pairs[np.ix_(~mask_valid_gt, (1, 3))] = -1.0  # type: ignore[reportArgumentType]
            # filter out invalid pd_id, pd_label_id, pd_score by setting to -1.0
            pairs[np.ix_(~mask_valid_pd, (2, 4, 6))] = -1.0  # type: ignore[reportArgumentType]
            # filter out invalid iou by setting to 0.0
            pairs[~mask_valid_pd | ~mask_valid_gt, 5] = 0.0

            for idx, col in enumerate(columns):
                column = pairs[:, idx]
                if col not in {"iou", "pd_score"}:
                    column = column.astype(np.int64)

                col_idx = tbl.schema.names.index(col)
                tbl = tbl.set_column(
                    col_idx, tbl.schema[col_idx], pa.array(column)
                )

            mask_invalid = ~mask_valid | (pairs[:, (1, 2)] < 0).all(axis=1)
            filtered_tbl = tbl.filter(pa.array(~mask_invalid))
            loader._detailed_writer.write_table(filtered_tbl)

        return loader.finalize(
            batch_size=batch_size,
            index_to_label_override=self._index_to_label,
        )

    def compute_precision_recall(
        self,
        iou_thresholds: list[float],
        score_thresholds: list[float],
        datums: pc.Expression | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all metrics except for ConfusionMatrix

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """
        if not iou_thresholds:
            raise ValueError("At least one IOU threshold must be passed.")
        elif not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        n_ious = len(iou_thresholds)
        n_scores = len(score_thresholds)
        n_labels = len(self._index_to_label)
        n_gts_per_lbl = extract_groundtruth_count_per_label(
            reader=self._detailed_reader,
            number_of_labels=n_labels,
            datums=datums,
        )

        counts = np.zeros((n_ious, n_scores, 3, n_labels), dtype=np.uint64)
        pr_curve = np.zeros((n_ious, n_labels, 101, 2), dtype=np.float64)
        running_counts = np.zeros((n_ious, n_labels, 2), dtype=np.uint64)

        for pairs in self._ranked_reader.iterate_arrays(
            numeric_columns=[
                "datum_id",
                "gt_id",
                "pd_id",
                "gt_label_id",
                "pd_label_id",
                "iou",
                "pd_score",
                "iou_prev",
            ],
            filter=datums,
        ):
            if pairs.size == 0:
                continue

            batch_counts = compute_counts(
                ranked_pairs=pairs,
                iou_thresholds=np.array(iou_thresholds),
                score_thresholds=np.array(score_thresholds),
                number_of_groundtruths_per_label=n_gts_per_lbl,
                number_of_labels=len(self._index_to_label),
                running_counts=running_counts,
                pr_curve=pr_curve,
            )
            counts += batch_counts

        # fn count
        counts[:, :, 2, :] = n_gts_per_lbl - counts[:, :, 0, :]

        precision_recall_f1 = compute_precision_recall_f1(
            counts=counts,
            number_of_groundtruths_per_label=n_gts_per_lbl,
        )
        (
            average_precision,
            mean_average_precision,
            pr_curve,
        ) = compute_average_precision(pr_curve=pr_curve)
        average_recall, mean_average_recall = compute_average_recall(
            prec_rec_f1=precision_recall_f1
        )

        return unpack_precision_recall_into_metric_lists(
            counts=counts,
            precision_recall_f1=precision_recall_f1,
            average_precision=average_precision,
            mean_average_precision=mean_average_precision,
            average_recall=average_recall,
            mean_average_recall=mean_average_recall,
            pr_curve=pr_curve,
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            index_to_label=self._index_to_label,
        )

    def compute_confusion_matrix(
        self,
        iou_thresholds: list[float],
        score_thresholds: list[float],
        datums: pc.Expression | None = None,
    ) -> list[Metric]:
        """
        Computes confusion matrices at various thresholds.

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        list[Metric]
            List of confusion matrices per threshold pair.
        """
        if not iou_thresholds:
            raise ValueError("At least one IOU threshold must be passed.")
        elif not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        n_ious = len(iou_thresholds)
        n_scores = len(score_thresholds)
        n_labels = len(self._index_to_label)

        confusion_matrices = np.zeros(
            (n_ious, n_scores, n_labels, n_labels), dtype=np.uint64
        )
        unmatched_groundtruths = np.zeros(
            (n_ious, n_scores, n_labels), dtype=np.uint64
        )
        unmatched_predictions = np.zeros_like(unmatched_groundtruths)

        for pairs in self._detailed_reader.iterate_arrays(
            numeric_columns=[
                "datum_id",
                "gt_id",
                "pd_id",
                "gt_label_id",
                "pd_label_id",
                "iou",
                "pd_score",
            ],
            filter=datums,
        ):
            if pairs.size == 0:
                continue

            (
                batch_mask_tp,
                batch_mask_fp_fn_misclf,
                batch_mask_fp_unmatched,
                batch_mask_fn_unmatched,
            ) = compute_pair_classifications(
                detailed_pairs=pairs,
                iou_thresholds=np.array(iou_thresholds),
                score_thresholds=np.array(score_thresholds),
            )
            (
                batch_confusion_matrices,
                batch_unmatched_groundtruths,
                batch_unmatched_predictions,
            ) = compute_confusion_matrix(
                detailed_pairs=pairs,
                mask_tp=batch_mask_tp,
                mask_fp_fn_misclf=batch_mask_fp_fn_misclf,
                mask_fp_unmatched=batch_mask_fp_unmatched,
                mask_fn_unmatched=batch_mask_fn_unmatched,
                number_of_labels=n_labels,
                iou_thresholds=np.array(iou_thresholds),
                score_thresholds=np.array(score_thresholds),
            )
            confusion_matrices += batch_confusion_matrices
            unmatched_groundtruths += batch_unmatched_groundtruths
            unmatched_predictions += batch_unmatched_predictions

        return unpack_confusion_matrix(
            confusion_matrices=confusion_matrices,
            unmatched_groundtruths=unmatched_groundtruths,
            unmatched_predictions=unmatched_predictions,
            index_to_label=self._index_to_label,
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
        )

    def compute_examples(
        self,
        iou_thresholds: list[float],
        score_thresholds: list[float],
        datums: pc.Expression | None = None,
    ) -> list[Metric]:
        """
        Computes examples at various thresholds.

        This function can use a lot of memory with larger or high density datasets. Please use it with filters.

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        list[Metric]
            List of confusion matrices per threshold pair.
        """
        if not iou_thresholds:
            raise ValueError("At least one IOU threshold must be passed.")
        elif not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        metrics = []
        tbl_columns = [
            "datum_uid",
            "gt_uid",
            "pd_uid",
        ]
        numeric_columns = [
            "datum_id",
            "gt_id",
            "pd_id",
            "gt_label_id",
            "pd_label_id",
            "iou",
            "pd_score",
        ]
        for tbl, pairs in self._detailed_reader.iterate_tables_with_arrays(
            columns=tbl_columns + numeric_columns,
            numeric_columns=numeric_columns,
            filter=datums,
        ):
            if pairs.size == 0:
                continue

            index_to_datum_id = {}
            index_to_groundtruth_id = {}
            index_to_prediction_id = {}

            # extract external identifiers
            index_to_datum_id = create_mapping(
                tbl, pairs, 0, "datum_id", "datum_uid"
            )
            index_to_groundtruth_id = create_mapping(
                tbl, pairs, 1, "gt_id", "gt_uid"
            )
            index_to_prediction_id = create_mapping(
                tbl, pairs, 2, "pd_id", "pd_uid"
            )

            (
                mask_tp,
                mask_fp_fn_misclf,
                mask_fp_unmatched,
                mask_fn_unmatched,
            ) = compute_pair_classifications(
                detailed_pairs=pairs,
                iou_thresholds=np.array(iou_thresholds),
                score_thresholds=np.array(score_thresholds),
            )

            mask_fn = mask_fp_fn_misclf | mask_fn_unmatched
            mask_fp = mask_fp_fn_misclf | mask_fp_unmatched

            batch_examples = unpack_examples(
                detailed_pairs=pairs,
                mask_tp=mask_tp,
                mask_fp=mask_fp,
                mask_fn=mask_fn,
                index_to_datum_id=index_to_datum_id,
                index_to_groundtruth_id=index_to_groundtruth_id,
                index_to_prediction_id=index_to_prediction_id,
                iou_thresholds=iou_thresholds,
                score_thresholds=score_thresholds,
            )
            metrics.extend(batch_examples)

        return metrics

    def compute_confusion_matrix_with_examples(
        self,
        iou_thresholds: list[float],
        score_thresholds: list[float],
        datums: pc.Expression | None = None,
    ) -> list[Metric]:
        """
        Computes confusion matrix with examples at various thresholds.

        This function can use a lot of memory with larger or high density datasets. Please use it with filters.

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        list[Metric]
            List of confusion matrices per threshold pair.
        """
        if not iou_thresholds:
            raise ValueError("At least one IOU threshold must be passed.")
        elif not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        metrics = {
            iou_idx: {
                score_idx: create_empty_confusion_matrix_with_examples(
                    iou_threhsold=iou_thresh,
                    score_threshold=score_thresh,
                    index_to_label=self._index_to_label,
                )
                for score_idx, score_thresh in enumerate(score_thresholds)
            }
            for iou_idx, iou_thresh in enumerate(iou_thresholds)
        }
        tbl_columns = [
            "datum_uid",
            "gt_uid",
            "pd_uid",
        ]
        numeric_columns = [
            "datum_id",
            "gt_id",
            "pd_id",
            "gt_label_id",
            "pd_label_id",
            "iou",
            "pd_score",
        ]
        for tbl, pairs in self._detailed_reader.iterate_tables_with_arrays(
            columns=tbl_columns + numeric_columns,
            numeric_columns=numeric_columns,
            filter=datums,
        ):
            if pairs.size == 0:
                continue

            index_to_datum_id = {}
            index_to_groundtruth_id = {}
            index_to_prediction_id = {}

            # extract external identifiers
            index_to_datum_id = create_mapping(
                tbl, pairs, 0, "datum_id", "datum_uid"
            )
            index_to_groundtruth_id = create_mapping(
                tbl, pairs, 1, "gt_id", "gt_uid"
            )
            index_to_prediction_id = create_mapping(
                tbl, pairs, 2, "pd_id", "pd_uid"
            )

            (
                mask_tp,
                mask_fp_fn_misclf,
                mask_fp_unmatched,
                mask_fn_unmatched,
            ) = compute_pair_classifications(
                detailed_pairs=pairs,
                iou_thresholds=np.array(iou_thresholds),
                score_thresholds=np.array(score_thresholds),
            )

            unpack_confusion_matrix_with_examples(
                metrics=metrics,
                detailed_pairs=pairs,
                mask_tp=mask_tp,
                mask_fp_fn_misclf=mask_fp_fn_misclf,
                mask_fp_unmatched=mask_fp_unmatched,
                mask_fn_unmatched=mask_fn_unmatched,
                index_to_datum_id=index_to_datum_id,
                index_to_groundtruth_id=index_to_groundtruth_id,
                index_to_prediction_id=index_to_prediction_id,
                index_to_label=self._index_to_label,
            )

        return [m for inner in metrics.values() for m in inner.values()]
